;*****************************************************************
;*****************************************************************
;                           UN32_8.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book NUMERICAL SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  January 25, 2001
; Last update:    
;
; Libary name: MATH32.LIB
; This module: UN32_8.ASM
; Operations : Finnancial calculations
;*****************************************************************
;
;*****************************************************************
;                           publics
;*****************************************************************
; MATHEMATICAL ROUTINES IN THIS MODULE
; Financial calculations
PUBLIC  _PAYMENTS        ; .. periodic payments on a loan
PUBLIC  _PRESENT_VALUE   ; .. present value of an investment
PUBLIC  _FUTURE_VALUE    ; .. future value of an investment
PUBLIC  _SL_DEPRECIATION ; .. straight-line depreciation
PUBLIC  _COMP_PERIODS    ; .. number of compounding periods
;
;****************************************************************
;               definitions for 32-bit flat model
;****************************************************************
        .486
        .MODEL flat
;*****************************************************************
;                           externals
;*****************************************************************
; In other modules of the MATH32 library
EXTRN   _X_TO_Y_BYBP: PROC   ;------ Un32_6 module 
EXTRN   _LOG_E:       PROC   ;----- in Un32_5 module  

        .DATA
;***************************|
;   Data for this module    |
;***************************|
; Data for financial procedures
PRINCIPAL       DT      0       ; Storage for principal
INTEREST        DT      0       ; Storage for interest rate
TERM            DT      0       ; Storage for term of loan
PAYMENT         DT      0       ; Storage for periodic payment
;****************************************************************
;                           code
;****************************************************************
        .CODE  
;******************************************
;            financial procedures
;******************************************
;
_PAYMENTS        PROC
; Financial procedure to calculate the ammount of each payment on
; a loan using the formula:
;
;                     i
;      P =  p * ---------------
;               1-(i + 1)**-n
; where
;     P = ammount of periodic payment
;     p = principal
;     i = periodic interest rate (in fractional form)
;     n = term of loan (number of payment periods)
;
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = p (principal)
; On exit:
;          ST(0) = P (ammount of periodic payment)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time, for example: to calculate monthly payments on a
;       3-year loan of $4000.00, at 0.14% yearly interest rate
;                   p = 4000
;                   i = 0.14/12 = 0.0116666...
;                   n = 3*12 = 36
;
;***************************|
;   store input data        |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   p   | ----- |
        FSTP    TERM            ;   i   |   p   | ----- |
        FSTP    INTEREST        ;   p   | ----- |
        FSTP    PRINCIPAL       ; ----- |
;***************************|
; calculate 1-(i + 1)**-n   |
;***************************|
        FLD     TERM            ;   n   | ----- |
        FLD     INTEREST        ;   i   |   n   | ----- |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   | ----- |
        FXCH                    ;   n   |  i+1  | ----- |
        CALL    _X_TO_Y_BYBP    ;(i+1)^n| ----- |
; x**-n = 1/x**n
        FLD1                    ;   1   |(i+1)^n| ----- |
        FDIVR                   ;  1/   |
                                ;(i+1)^n| ----- |
        FLD1                    ;  1    |1/(i+..| ----- |
        FSUBR                   ;1-(1/..| ----- |
;***************************|
; calculate:                |
;     p * i/1-(i + 1)**-n   |
;***************************|
; den = 1-(i + 1)**-n
;                               ; den   | ----- |
;        FLD1                    ;  1    |  den  | ----- |
;        FSUBR                   ; 1-den | ----- |
        FLD     INTEREST        ;   i   | 1-den | ----- |
        FDIVR                   ;i/1-den| ----- |
        FLD     PRINCIPAL       ;  p    |i/1-den| ----- |
        FMULP   ST(1),ST        ;  P    | ----- |
        CLD
        RET
_PAYMENTS        ENDP
;
;
_PRESENT_VALUE   PROC
; Financial procedure to calculate the present value of an
; investment using the formula:
;
;                 1-(i + 1)**-n
;      PV =  pm * ---------------
;                       i
;
; where:
;     PV = present value of an investment
;     pm = periodic payment
;     i = periodic interest rate (in fractional form)
;     n = term of loan (number of payment periods)
;
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = pm (periodic payment)
; On exit:
;          ST(0) = PV (present value of investment)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time (see note in the PAYMENTS procedure)
;
;***************************|
;   store input data        |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   pm  | ----- |
        FSTP    TERM            ;   i   |   pm  | ----- |
        FSTP    INTEREST        ;   pm  | ----- |
        FSTP    PAYMENT         ; ----- |
;***************************|
; calculate 1-(i + 1)**-n   |
;***************************|
        FLD     TERM            ;   n   | ----- |
        FLD     INTEREST        ;   i   |   n   | ----- |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   | ----- |
        FXCH                    ;   n   |  i+1  | ----- |
        CALL    _X_TO_Y_BYBP    ;(i+1)^n| ----- |
; x**-n = 1/x**n
        FLD1                    ;   1   |(i+1)^n| ----- |
        FDIVR                   ;  1/   |
                                ;(i+1)^n| ----- |
        FLD1                    ;  1    |1/(i+..| ----- |
        FSUBR                   ;1-(1/..| ----- |
;***************************|
; calculate:                |
;           1-(i + 1)**-n   |
;   pm *  ---------------   |
;                i          |
;***************************|
; num = 1-(i + 1)**-n
;                               ; num   | ----- |
        FLD     INTEREST        ;   i   |  num  | ----- |
        FDIV                    ; num/i | ----- |
        FLD     PAYMENT         ;  pm   | i/num | ----- |
        FMULP   ST(1),ST        ;  PV   | ----- |
        CLD
        RET
_PRESENT_VALUE   ENDP
;
;
_FUTURE_VALUE    PROC
; Financial procedure to calculate the future value of an
; investment using the formula:
;
;                 (i + 1)**n - 1
;      FV =  pm * ---------------
;                       i
;
; where:
;     PV = present value of an investment
;     pm = periodic payment
;     i = periodic interest rate (in fractional form)
;     n = term of loan (number of payment periods)
;
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = pm (periodic payment)
; On exit:
;          ST(0) = FV (future value of investment)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time (see note in the PAYMENTS procedure)
;
;***************************|
;   store input data        |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   pm  | ----- |
        FSTP    TERM            ;   i   |   pm  | ----- |
        FSTP    INTEREST        ;   pm  | ----- |
        FSTP    PAYMENT         ; ----- |
;***************************|
; calculate (i + 1)**n - 1  |
;***************************|
        FLD     TERM            ;   n   | ----- |
        FLD     INTEREST        ;   i   |   n   | ----- |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   | ----- |
        FXCH                    ;   n   |  i+1  | ----- |
        CALL    _X_TO_Y_BYBP    ;(i+1)^n| ----- |
        FLD1                    ;   1   |(i+1)^n| ----- |
        FSUB                    ;(i+1)^n|
                                ;  -1   | ----- |
;***************************|
; calculate:                |
;          (i + 1)**n - 1   |
;   pm *  ---------------   |
;                i          |
;***************************|
; num = (i + 1)**n - 1
;                               ; num   | ----- |
        FLD     INTEREST        ;   i   |  num  | ----- |
        FDIV                    ; num/i | ----- |
        FLD     PAYMENT         ;  pm   | i/num | ----- |
        FMULP   ST(1),ST        ;  PV   | ----- |
        CLD
        RET
_FUTURE_VALUE    ENDP
;
;
_SL_DEPRECIATION PROC
; Financial procedure to calculate the straight-line depreciation
; of an asset during 1 period, using the formula:
;
;              (c - s)
;      SLD =  --------
;                 ul
; where:
;     SLD = straight-line depreciation allowance
;       c = initial cost of the asset
;       s = salvage value of the asset at the end of the
;           depreciation period
;      ul = useful life of the asset
;
; On entry:
;          ST(0) = ul (useful life)
;          ST(1) = s (salvage value)
;          ST(2) = c (initial cost)
; On exit:
;          ST(0) = SLD allowance per period
;
; Note: the depreciation period (SLD) and the useful life (ul) of
;       asset must be in the same units of time
;
;***************************|
;   calculate (c-s)/ul      |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   ul  |   s   |   c   | ----- |
        FXCH    ST(2)           ;   c   |   s   |   ul  | ----- |
        FSUB    ST,ST(1)        ;  c-s  |   s   |   ul  | ----- |
        FSTP    ST(1)           ;  c-s  |   ul  | ----- |
        FDIV    ST,ST(1)        ; c-s/ul|   ul  | ----- |
        FSTP    ST(1)           ;  SLD  | ----- |
        CLD
        RET
_SL_DEPRECIATION ENDP
;
;
_COMP_PERIODS    PROC
; Financial procedure to calculate the number of compounding
; periods necessary for an investment to grow to a predetermined
; value, by means of the formula:
;
;           ln(fv/sv)
;      CP = ---------
;           ln(1 + i)
;
; where:
;      CP = number of compounding periods
;      fv = final value
;      sv = starting value
;       i = interest rate
;      ln = log base e (natural logarithm)
; On entry:
;          ST(0) = i (interest rate)
;          ST(1) = fv (final value of investment)
;          ST(2) = sv (starting value of investment)
; On exit:
;          ST(0) = CP (number of compounding periods for growth)
;
; Note: the number of compounding periods (CP) will be in the same
;       units of time as the interest rate (i). For example: to
;       calculate compounding periods (in months) for an investment
;       at a yearly interest of 14 percent
;                 i = 0.14/12 = 0.0116666...
;
;***************************|
;   calculate ln(1 + i)     |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   i   |   fv  |   sv  | ----- |
        FLD1                    ;   1   |   i   |   fv  |   sv  |
        FADD                    ; 1 + i |   fv  |   sv  | ----- |
        CALL    _LOG_E          ;ln(1+i)|   fv  |   sv  | ----- |
;***************************|
;   calculate:
;      ln(fv/sv)/ln(1 + i)  |
;***************************|
        FXCH    ST(2)           ;  sv   |   fv  |ln(1+i)| ----- |
        FDIV                    ; fv/sv |ln(1+i)| ----- |
        CALL    _LOG_E          ;ln(fv/ |
                                ;  sv)  |ln(1+i)| ----- |
        FDIVR                   ;  CP   | ----- |
        CLD 
        RET
_COMP_PERIODS    ENDP

        end


    