//************************************************************
//               Financial functions
//************************************************************
// Reference:   NUMERICAL SOLUTIONS
// Workspace  : Chapter9\Test Un32_8
// Entry point: Test Un32_8
// Other files: Un32_8.asm
//              Un32_8.h
//              Un32_8.obj
//              Un32_6.obj
//              Un32_5.obj      
// Development: Visual C++ 6.0 Console Application
// Date:        January 25/2001
//
// Description:
//              Header file for testing calculator primtive
//              functions in the Un32_8 module of math32.lib
//************************************************************
// Library functions
extern "C" PAYMENTS();       //..  periodic payments on a loan
extern "C" ANNUITY_PV();     // .. present value of an investment
extern "C" ANNUITY_FV();     //..  future value of an investment
extern "C" SL_DEPRECIATION();// .. straight-line depreciation
extern "C" COMP_PERIODS();   // .. number of compounding periods
extern "C" SIMPLE_INTEREST();// .. simple interest
extern "C" FUTURE_VALUE();   // .. future value at comp. int.
extern "C" PRESENT_VALUE();  // .. present value at comp. int.
extern "C" EFFECTIVE_YIELD();// .. effective annual yield
extern "C" APR();            // .. annual percentage rate
extern "C" AOI_PAYMENT();    // .. add-on interest payment
extern "C" ANNUITY_DUE();    // .. annuity at start of period
extern "C" SINKING_FUND();   // .. periodic payment in annuity
//**************************************************************
//                       Prototypes
//**************************************************************
double Payments(double, double, double, int);
double AnnuityPv(double, double, double, int);
double AnnuityFv(double, double, double, int);
double SlDepreciation(double, double, int, int);
double CompPeriods(double, double, double, int);
double FutureValue(double, double, double, int);
double PresentValue(double, double, double, int);
double EffectiveYield(double, int);
double Apr(double, int);
double AoiPayment(double, double, double, int);
double AnnuityDue(double, double, double, int);
double SinkingFund(double, double, double, int);

//**************************************************************
//            C++ interface functions to Un32_8 module
//**************************************************************
double SimpleInterest(double term, double i, double p)
{
// Simple interest calculation according to the formula:
//
//          I = Prt
// where:
//          P = principal amount (present value)
//          r = Annual interest rate
//          t = Time in years
//          I = Amount of interest
// On entry:
//          parameter term = time in years
//          parameter i = interest rate
//          parameter p = principal 
// On exit:
//          returns interest amount
double simpInt;
     // Convert interest to per-cent form
     i /= 100;
_asm
{
    FLD      term      ; Load years
	FLD      p         ; Load principal
	FLD      i         ; Interest
	;                     ST0     ST1     ST2
	; ASSERT:          |   i   |   n   |   p   |
    CALL     SIMPLE_INTEREST
	FSTP     simpInt   ;|   I  |
}
return simpInt;
}


double FutureValue(double term, double i, double p, int ppy)
{
// Finnacial procedure to calculate the future value of an investment
// at a compound interest, according to the formula:
//
//         A = P(i + 1)^n
//
// where
//         P = present value
//         i = interest rate per compounding period
//         n = number of compunding periods
//         A = Value of investment at end of period
// On entry:
//          parameter term in years
//          parameter i = interest rate
//          parameter p = present value
// On exit:
//          returns future value of investment
//
// Note: interest rate (i) and term (n) must be in the same units
//       of time, for example: to calculate monthly compounding periods
//       on a 3-year loan, at 12% yearly interest rate
//                   i = 0.12/12 = 0.01
//                   n = 3*12 = 36     
//       Preliminary code must perform these adjustments
double futVal;
     // Convert interest to per-cent form
     i /= 100;	
_asm
{
	FLD      p         ; Load present value
	FILD     ppy       ; Time units
	FLD      i         ; Interest
	FDIV     ST,ST(1)  ; i/t
	FLD      term      ; term
	FMUL     ST,ST(2)  ; term*t
	FSTP     ST(2)
	FXCH
	;                     ST0     ST1     ST2
	; ASSERT:          |   n   |   i   |   p   |
    CALL     FUTURE_VALUE
	FSTP     futVal
}
return futVal;
}

double PresentValue(double term, double i, double a, int ppy)
{
// Finnacial procedure to calculate the present value of an investment
// at compound interest, according to the formula:
//
//         P = A(i + 1)^-n
//
// where
//         P = present value
//         i = interest rate per compounding period
//         n = number of compunding periods
//         A = Value of investment at end of period
// On entry:
//          parameter term in years
//          parameter i = interest rate
//          parameter a = future value
// On exit:
//          returns present value of investment
//
// See note regarding time units in the FutureValue() function.

double presVal;
     // Convert interest to per-cent form
     i /= 100;	
_asm
{
	FLD      a         ; Load future value
	FILD     ppy       ; Time units
	FLD      i         ; Interest
	FDIV     ST,ST(1)  ; i/t
	FLD      term      ; years
	FMUL     ST,ST(2)  ; n*t
	FSTP     ST(2)
	FXCH
	;                     ST0     ST1     ST2
	; ASSERT:          |   n   |   i   |   p   |
    CALL     PRESENT_VALUE
	FSTP     presVal
}
return presVal;
}

double EffectiveYield(double i, int ppy)
{
// Finnacial procedure to calculate the effective annual yield
// an investment at compound interest, according to the formula:
//
//         Y = ((1 + i/n)^n) - 1
//
// where
//         i = interest rate per compounding period
//         n = number of compunding periods
//         Y = effective annual yield of investment
// On entry:
//         parameter i = interest rate
//         parameter n = term
// On exit:
//         returns effective annual yield of investment
//
// See note regarding time units in the FutureValue() function.

double yield;
     // Convert interest to per-cent form
     i /= 100;	
_asm
{
	FILD     ppy       ; Periods per year = n
	FLD      i         ; Interest
                       ;  ST0  |  ST1  |  ST2
	                   ;   i   |   n   | ----- | 
	; ASSERT:          |   i   |   n   | ----- |
    CALL     EFFECTIVE_YIELD
	FSTP     yield
}

return yield * 100;     // Convert to per-cent
}

double Apr(double i, int np)
{
// Financial procedure to calculate the annual percentage rate
// APR, according to the formula:
//
//              2rN
//      APR = -------
//             N + 1
//
// where
//         N = number of payments
//         r = add-on interest rate
//       APR = annual percentage rate
// On entry:
//         parameter i = add-on interest
//         parameter np = number of payments
// On exit:
//         returns APR (annual percentage rate)
double apr;
     // Convert interest to per-cent form
     i /= 100;	
_asm
{
	FILD     np        ; Number of payments
	FLD      i         ; Add-on interest
                       ;  ST0  |  ST1  |  ST2
	                   ;   i   |   N   | ----- | 
    CALL     APR
	FSTP     apr
}
return apr*100;     // Convert to per-cent
}

double AoiPayment(double t, double i, double p, int ppy)
{
// Finnacial procedure to calculate the individual payments
// on a add-on interest loan, according to the formula:
//
//              Pit
//      AOI =  -------
//               N
//
// where
//         N = number of payments
//         i = add-on interest rate
//         t = loan time in years
//         P = principal
//       AOI = amount of each add-on intrest payment
// On entry:
//         parameter = i (add-on interest)
//         parameter = p (principal)
//         parameter ppy = (number of payments per year)
//         parameter t = t (loan time in years)
// On exit:
//         Returns amount of each payment
double aoi;
     // Convert interest to per-cent form
     i /= 100;
_asm
{
;           N = t*ppy = 3*12 = 36
; AOI_PAYMENT expects:
;                         STO     ST1     ST2     ST3
;                      |   p   |   i   |   t   |   N   |
; N = ppy * t
;       
;                         STO      ST1     ST2     ST3
    FILD     ppy       ;  ppy   | ----- |
    FLD      t         ;   t    |  ppy  | ----- |
    FMUL     ST,ST(1)  ;   N    |  ppy  | ----- |
    FSTP     ST(1)     ;   N    | ----- |
    FLD      t         ;   t    |   N   | ----- |
    FLD      i         ;   i    |   t   |   N   | ----- |
	FLD      p         ;   p    |   i   |   t   |   N   |

; ASSERT:              ;   p    |   i   |   t   |   N   |
    CALL     AOI_PAYMENT
	FSTP     aoi
}
return aoi;
}


double Payments(double term, double i, double p, int ppy)
{
// Financial procedure to calculate the amount of each payment on
// a loan using the formula:
//
//                     i
//      P =  p * ---------------
//               1-(i + 1)^-n
// where
//     P = ammount of periodic payment
//     p = principal (present value)
//     i = interest rate (in fractional form)
//     n = term of loan (number of payment periods)
//
// On entry:
//          parameter term = years
//          parameter i = interest rate
//          parameter p = principal
//          parameter ppy = payment periods per year	 
// On exit:
//          returns ammount of periodic payment
//
double prin;

     // Convert interest to per-cent form
     i /= 100;
_asm
{
; Interest rate (i) and term (n) are converted to the same
; same units of time before calling the PAYMENT function
;    For example: to calculate monthly payments on a 3-year
;    loan of $4000.00, at 0.14% yearly
;                   p = 4000
;                   i = 0.14
;                   n = 3
;                 ppy = 12
;               i/ppy = 0.14/12 = 0.0116666...
;               n*ppy = 3*12    = 3
	FLD      p         ; Load principal
	FILD     ppy       ; Time units
	FLD      i         ; Interest
	FDIV     ST,ST(1)  ; i/t
	FLD      term      ; years
	FMUL     ST,ST(2)  ; n*t
	FSTP     ST(2)
	FXCH
	;                     ST0     ST1     ST2
	; ASSERT:          |   n   |   i   |   p   |
    CALL     PAYMENTS
	FSTP     prin     ;|   P   |
}
return prin;
}

double AnnuityPv(double term, double i, double pm, int ppy)
{
// Financial procedure to calculate the present value of an
// investment using the formula:
//
//                 1-(i + 1)^-n
//      PV =  pm * ---------------
//                       i
//
// where:
//     PV = present value of an investment
//     pm = periodic payment
//     i = periodic interest rate 
//     n = term of loan in years
//
// On entry:
//          parameter term = years
//          parameter i = interest rate
//          parameter pm  = periodic payment
//          parameter ppy = payment periods per year	 
// On exit:
//          returns present value of investment
double pValue;

     // Convert interest to per-cent form
     i /= 100;
_asm
{
; Interest rate (i) and term (n) are converted to the same
; same units of time before calling the PRESENT_VALUE
; function
	FLD      pm        ; Load periodic payment
	FILD     ppy       ; Time units
	FLD      i         ; Interest
	FDIV     ST,ST(1)  ; i/t
	FLD      term      ; years
	FMUL     ST,ST(2)  ; n*t
	FSTP     ST(2)
	FXCH
	;                     ST0     ST1     ST2
	; ASSERT:          |   n   |   i   |   pm  |
    CALL     ANNUITY_PV
	FSTP     pValue
}
return pValue;
}

double AnnuityFv(double term, double i, double pm, int ppy)
{
// Financial procedure to calculate the future value of an
// investment using the formula:
//
//                 (i + 1)^(n-1)
//      FV =  pm * ---------------
//                       i
//
// where:
//     FV = future value of an investment
//     pm = periodic payment
//     i = periodic interest rate 
//     term = years
//
// On entry:
//          parameter term = years
//          parameter i = interest rate
//          parameter pm  = periodic payment
//          parameter ppy = payment periods per year	 
// On exit:
//          returns future value of investment
double fValue;

     // Convert interest to per-cent form
     i /= 100;
_asm
{
; Interest rate (i) and term (n) are converted to the same
; same units of time before calling the FUTURE_VALUE function
	FLD      pm        ; Load periodic payment
	FILD     ppy       ; Time units
	FLD      i         ; Interest
	FDIV     ST,ST(1)  ; i/t
	FLD      term      ; years
	FMUL     ST,ST(2)  ; n*t
	FSTP     ST(2)
	FXCH
	;                     ST0     ST1     ST2
	; ASSERT:          |   n   |   i   |   pm  |
    CALL     ANNUITY_FV
	FSTP     fValue
}
return fValue;
}


double AnnuityDue(double term, double i, double pm, int ppy)
{
// Financial procedure to calculate the future value of an
// annuity due, using the formula:
//
//                 (i + 1)^(n+1) - 1
//      A =  pm * ------------------ - 1
//                         i
//
// where:
//     A = annuity due at the beginning of each period
//     pm = periodic payment
//     i = periodic interest rate (in fractional form)
//     n = number of payment periods
//
// On entry:
//          parameter term is years
//          parameter i is interest rate
//          parameter pm is periodic payment
//          paramenter ppy is payments per year
// On exit:
//          Returns annuity due
//
// Note: interest rate (i) and term (n) must be in the same units
//       of time (see note in the PAYMENTS procedure)

double aDue;

     // Convert interest to per-cent form
     i /= 100;
_asm
{
; Interest rate (i) and term (n) are converted to the same
; same units of time before calling the ANNUITY_DUE procedure
	FLD      pm        ; Load periodic payment
	FILD     ppy       ; Time units
	FLD      i         ; Interest
	FDIV     ST,ST(1)  ; i/t
	FLD      term      ; years
	FMUL     ST,ST(2)  ; n*t
	FSTP     ST(2)
	FXCH
	;                     ST0     ST1     ST2
	; ASSERT:          |   n   |   i   |   pm  |
    CALL     ANNUITY_DUE
	FSTP     aDue
}
return aDue;
}

double SinkingFund(double term, double i, double a, int ppy)
{
// Financial procedure to calculate the periodic payment
// (sinking fund) required in an annuity, using the formula:
//
//                Ai   
//     pm =  --------------
//           (i + 1)^n - 1
//
// where:
//     pm = periodic payment
//     i = periodic interest rate (in fractional form)
//     n = term of loan (number of payment periods)
//     A = annuity
//
// On entry:
//          parameter term = n          --> ST(0)
//          parameter i = interest rate --> ST(1)
//          parameter a = A             --> ST(2)
// On exit:
//          returns pm (periodic payment required)
//
// Note: interest rate (i) and term (n) must be in the same units
//       of time (see note in the PAYMENTS procedure)
double pm;

     // Convert interest to per-cent form
     i /= 100;
_asm
{
; Interest rate (i) and term (n) are converted to the same
; same units of time before calling the SINKING_FUND procedure
	FLD      a         ; Load future value of annuity
	FILD     ppy       ; Time units
	FLD      i         ; Interest
	FDIV     ST,ST(1)  ; i/t
;                      |  ST0  |  ST1  |  ST2  |  ST3  |                       
	                   ; i/t   |   ppy |   A   | ----- |  
	FLD      term      ;  n    |   i/t |  ppy  |   A   |
	FMUL     ST,ST(2)  ; n*t   |   i/t |  ppy  |   A   | 
	FSTP     ST(2)     ; n*t   |   i/t |   A   | ----- |
	FXCH
	;                     ST0     ST1     ST2
	; ASSERT:          |   n   |   i   |   A   |
    CALL     SINKING_FUND
	FSTP     pm
}
return pm;
}


double SlDepreciation(double cost, double sValue, double ul, int dpy)
{
// Financial procedure to calculate the straight-line depreciation
// of an asset during one period, using the formula:
//
//              (c - s)
//      SLD =  --------
//                ul
// where:
//     SLD = straight-line depreciation allowance
//       c = initial cost of the asset
//       s = salvage value of the asset at the end of the
//           depreciation period
//      ul = useful life of the asset
//
// On entry:
//      parameter cost = initial cost
//      parameter sValue = salvage value of asset
//      parameter ul = useful life in years
//      parameter dpy = depreciation periods per year
// On exit:
//      returns straight line depreciation allowance
//      per depreciation period
//
double sld;
_asm
{
; Interest rate (i) and term (n) are converted to the same
; same units of time before calling the primtive function
	FLD      cost      ; Load cost
	FLD      sValue    ; Load salvage value
	FILD     dpy       ; Depreciation time units
	FLD      ul        ; Useful life in years
	FMUL     ST,ST(1)  ; ul * dpy
	FSTP     ST(1)
	;                     ST0     ST1     ST2
	; ASSERT:          |   n   |  sal  | cost |
    CALL     SL_DEPRECIATION
	FSTP     sld
}
return sld;
}

double CompPeriods(double i, double fv, double sv, int cpy)
{
// Financial procedure to calculate the number of compounding
// periods necessary for an investment to grow to a predetermined
// value, by means of the formula:
//
//           ln(fv/sv)
//      CP = ---------
//           ln(1 + i)
//
// where:
//      CP = number of compounding periods
//      fv = final value
//      sv = starting value
//       i = interest rate
//      ln = log base e (natural logarithm)
// On entry:
//      parameter i = interest rate
//      parameter fv = final value
//      parameter sv = start value
//      parameter cpy = compounding periods per year
// On exit:
//      returns number of compunding period required
//
double cp;

// Convert interest to per-cent form
     i /= 100;
_asm
{
	FLD      sv        ; Start value
    FLD      fv        ; Final value
	FILD     cpy       ; Time units
	FLD      i         ; Interest
	FDIV     ST,ST(1)  ; i / cpy
	FSTP     ST(1)
	;                     ST0     ST1     ST2
	; ASSERT:          |   i   |  fv   |  sv   |
    CALL     COMP_PERIODS
	FSTP     cp
}
return cp;
}


     
	