;*****************************************************************
;*****************************************************************
;                           UNIT_2.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  January 13, 1994
; Updated:        January 25, 1994
;
; Libary name: MATHUNIT.LIB
;
;*****************************************************************
;                           publics
;*****************************************************************
; MATHEMATICAL ROUTINES IN THIS MODULE
PUBLIC  SIGN_ADD_BCD12  ; .. Signed BCD adition
PUBLIC  SIGN_SUB_BCD12  ; .. Signed BCD subtraction
PUBLIC  ID_CPU          ; .. Indentify processor
PUBLIC  ID_FPU          ; .. Identify floating-point unit
PUBLIC  NUM_AT_ST0      ; .. Identify value at ST(0)
;
;*****************************************************************
;                           externals
;*****************************************************************
; Routines in other modules of MATHUNIT.LIB
EXTRN   BCD_INTO_REG:   FAR
EXTRN   BIN_TO_BCD:     FAR
;****************************************************************
;                          code segment
;****************************************************************
;
CODE    SEGMENT BYTE
        ASSUME  CS:CODE
;***************************|
;   code segment data for   |
; procedures in this module |
;***************************|
; Scratch pad area for routine work variables
ENTRY_SI        DW      0       ; Storage for entry pointer SI
ENTRY_DI        DW      0       ; Storage for entry pointer DI
ENTRY_BX        DW      0       ; Storage for entry pointer BX
; Biased exponents
BIAS_X          DW      0       ; Biased exponent for x
BIAS_Y          DW      0       ; Biased exponent for y
BIAS_Z          DW      0       ; Biased exponent for sum
; Work areas for significand operations
SIG_L   DB      36 DUP (00H)    ; Larger significand area
SIG_S   DB      36 DUP (00H)    ; Smaller significand area
SIG_R   DB      37 DUP (00H)    ; Significand of result
                                ; Extra byte holds possible carry
; Others
LARGER  DB      0               ; 'x' or 'y', whichever is larger
CASE_12 DB      0               ; Switch for CASE 1 or CASE 2
; Partial products buffer for multiplication operations
PART_PROD       DB      37 DUP (00H)
                                ; Buffer for partial products
;*******************************************************************
;                     BCD arithmetic routines
;*******************************************************************
;************************************|
; signed floating point BCD addition |
;************************************|
SIGN_ADD_BCD12  PROC    FAR
; Addition of two signed BCD numbers stored in BCD12 floating point
; format
; Operation:
;           z = x + y
;           where x, y, and z are signed, floating point numbers
; On entry:
;          DS:SI --> Addend (element x in z = x + y)
;          DS:DI --> Augend (element y in z = x + y)
;          DS:BX --> 12-byte storage area for result in BCD12 format
; Note: the code assumes that the BCD12 numbers are in normalized
;       form, that is, that there are no leading zeros in the
;       significand
; On exit:
;       DS:BX --> Sum (element z in z = x + y)
;       Entry pointers are preserved
;
; This routine operates on two numbers encoded in BCD12 format (see
; Figure 2.10) as follows:
;
;          Ss e e e e m . m m m m m m m m m m m m m m m m m m
;                       |
;                       |____ implicit decimal point
;
;    S = sign of number (1 BCD digit)
;    s = sign of exponent (1 BCD digit)
;    e = exponent (4 BCD digits)
;    m = normalized significand (18 BCD digits)
;        (first significand digit must be non-zero)
;    . = implicit decimal point between the first and second
;        significand digits
;*******************************************************************
; BCD signed addition algorithm:
; CASE 1:
;    If x and y have the same sign, the absolute values are
;    added and the result has the common sign
; CASE 2:
;    If x and y have different signs, the smaller value is
;    subtracted from the larger value and the result has the sign
;    of the larger
;*******************************************************************
; Routine operations
; CASE 1 and 2:
; A. The input elements are tested for zero values. If one element
;    is zero the result is the value of the other element
; B. The packed siginificands in BCD12 format are unpacked and moved
;    into work buffers located in the code segment
; C. The upacked significands are aligned in the work buffers SIG_L
;    (for the significand of the number with the larger absolute
;     value) and SIG_S (for the significand of the number with the
;     smaller absolute value)
; CASE 1 (x and y have the same sign)
;     SIG_R will hold the significand of the sum
;     Addition operation:
;      SIG_L =    D D D . . . D D D 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
;  +   SIG_S =    0 0 0 0 d d d . . . d d d 0 0 0 0 0 0 0 0 0 0 0
;               -------------------------------------------------
;      SIG_R =  C s s s s s s s s s s s s s 0 0 0 0 0 0 0 0 0 0 0
;          leyend:
;             D = digits in the larger significand
;             d = digits in the smaller sginificand
;             s = digits in the sum
;             C = possible carry digit in the sum significand
;                 (in this case the exponent must be adjusted)
; CASE 2 (x and y have different signs)
;     SIG_R will hold the significand of the difference
;     Subtraction operation:
;      SIG_L =    D D D . . . D D D 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
;  -   SIG_S =    0 0 0 0 d d d . . . d d d 0 0 0 0 0 0 0 0 0 0 0
;               -------------------------------------------------
;      SIG_R =  B s s s s s s s s s s s s s 0 0 0 0 0 0 0 0 0 0 0
;          leyend:
;             B = possible borrow digit in the difference
; TRIVIAL CASE:
;    If the difference between exponents is larger than the number
;    of significand digits, then the aligned significands will be
;    as follows:
;      SIG_L =    D D D . . . D D D 0 0 0 0 0 0 0 0 0 0
;      SIG_S =    0 0 0 0 0 0 0 0 0 0 d d d . . . d d d
;              ----------------------------------------
;      SIG_R =  0 D D D . . . D D D 0 ? ? ? . . . ? ? ?
;    This means that the result (rounded to the format's significand
;    size) will equal the significand of the larger number. Therefore
;    the addition or subtraction operation would be trivial
; D. The exponent of the sum is the exponent of the element with
;    the larger absolute value, adjusted according to the operations
;    performed on the significands
;*******************************************************************
; Store entry variables and clear buffers
        CALL    STORE_AND_CLEAR         ; Local procedure
;**********************|
;    test x or y = 0   |
;**********************|
; If x or y = 0 then the sum is the non-zero element
        CALL    TEST_4_ZERO     ; Test routine
        JNC     X_NOT_ZERO      ; Go if no carry
; At this point x is zero
        MOV     SI,DI           ; Set SI to point to y
MOVE_2Z:
        MOV     CX,12           ; Count for digits to move
MOVE_12_2Z:
        MOV     AL,BYTE PTR[SI] ; Get byte from source
        MOV     BYTE PTR[BX],AL ; and store by BX pointer
        INC     SI              ; Bump pointers
        INC     BX
        LOOP    MOVE_12_2Z
;**********************************|
;           exit routine           |
;**********************************|
SIGN_ADD_EXIT:
        MOV     SI,CS:ENTRY_SI  ;Restore entry pointers
        MOV     DI,CS:ENTRY_DI
        MOV     BX,CS:ENTRY_BX
        RET                     ; End of processing
;**********************|
;    test for y = 0    |
;**********************|
X_NOT_ZERO:
        PUSH    SI              ; Save x pointer
        MOV     SI,DI           ; Pointer to SI
        CALL    TEST_4_ZERO     ; Test routine
        POP     SI              ; Restore x pointer
        JNC     XY_NOT_ZERO     ; Continue if no carry
        JMP     MOVE_2Z         ; Exit routine if y = 0
;**********************|
;  neither x nor y = 0 |
;**********************|
; Convert BCD signed exponent to biased binary exponent
; Binary bias = 10000
; 0, 1, 2 ............... 10000 ...............9997, 9998, 9999
;  negative exponents  <-- zero -->  positive exponents
;**********************|
;  exponent to biased  |
;      10000 form      |
;**********************|
; First convert exponent of x
; SI points to x
XY_NOT_ZERO:
        CALL    EXP_2_BIAS      ; Conversion procedure
; DX now holds binary exponent for x in bias 10,000 form
        MOV     CS:BIAS_X,DX    ; Store in scratch pad area
; Obtain biased exponent for y
        XCHG    SI,DI           ; Save x and set y
        CALL    EXP_2_BIAS      ; Conversion procedure
        MOV     CS:BIAS_Y,DX    ; Store in scratch pad area
        XCHG    SI,DI           ; Restore x and y pointers
;**********************|
;   determine which    |
;   number is larger   |
;**********************|
; DX = biased exponent for y
        CMP     CS:BIAS_X,DX    ; Compare exponent of x and y
        JA      X_BIGGER        ; x exponent is larger
        JB      Y_BIGGER        ; y exponent is larger
;**********************|
; compare significands |
;**********************|
; At this point x and y have the same exponent
; To determine which number is larger or if they are equal it is
; necessary to look at the digits in the significands
; SI --> x and DI --> y
        ADD     SI,3            ; Pointer to significand digits
        ADD     DI,3
        MOV     CX,9            ; 9 packed BCD digits to compare
COMPARE_SIGS:
        MOV     AL,[SI]         ; Get two significand digits of x
        CMP     AL,BYTE PTR [DI]        ; Compare to two y digits
        JA      X_BIGGER        ; Go if x > y
        JB      Y_BIGGER        ; or if y > x
; At this point the compared digits are equal
        INC     SI              ; Bump pointers
        INC     DI
        LOOP    COMPARE_SIGS    ; Continue comparing digits
;**********************|
;       x = y          |
;**********************|
; At this point x and y are identical numbers
        MOV     CS:LARGER,'e'   ; Code for x = y
        JMP     X_EQUAL_Y       ; Go to entry point
;**********************|
; y exponent is larger |
;**********************|
Y_BIGGER:
        MOV     CS:LARGER,'y'   ; y is larger than x
; Move 18 packed BCD digits from BCD12 encoding to SIG_L buffer
X_EQUAL_Y:
        MOV     SI,CS:ENTRY_DI  ; Offet of y value pointer
        ADD     SI,3            ; Mantissa digits in BCD12 encoding
        LEA     DI,CS:SIG_L     ; Offset of CS buffer
        MOV     CX,9            ; 9 packed digits to move
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
        MOV     SI,CS:ENTRY_SI  ; Set pointer to smaller number
; Get exponents
        MOV     DX,CS:BIAS_Y    ; Load y exponent
        MOV     CS:BIAS_Z,DX    ; and store as exponent of sum
        MOV     AX,CS:BIAS_X    ; Smaller exponent to AX
        JMP     X_OR_Y          ; Go to next routine
;**********************|
; x exponent is larger |
;**********************|
X_BIGGER:
        MOV     CS:LARGER,'x'   ; x is larger than y
; Move 18 packed BCD digits from BCD12 encoding to SIG_L buffer
        MOV     SI,CS:ENTRY_SI  ; Offet of x value pointer
        ADD     SI,3            ; Mantissa digits in BCD12 encoding
        LEA     DI,CS:SIG_L     ; Offset of CS buffer
        MOV     CX,9            ; 9 packed digits to move
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
        MOV     SI,CS:ENTRY_DI  ; Set pointer to smaller number
; Get larger exponent
        MOV     DX,CS:BIAS_X    ; Load x exponent
        MOV     CS:BIAS_Z,DX    ; and store as exponent of sum
        MOV     AX,CS:BIAS_Y    ; Smaller exponent to AX
X_OR_Y:
; At this point DX holds the larger exponent and AX the smaller one
; Test for a difference that is larger than the number of digits
; in the significand (18 digits in the BCD12 format)
        SUB     DX,AX           ; Larger minus smaller
        CMP     DX,18           ; Test for limit
        JBE     NOT_TRIVIAL     ; Not trivial case if <= 18
;**********************|
;  trivial operation   |
;**********************|
; If one element is too small to affect the sum, then the result is
; the larger element
; Assume that x > y
        MOV     SI,CS:ENTRY_SI  ; x value to SI
        MOV     BX,CS:ENTRY_BX
        CMP     CS:LARGER,'x'   ; Test if x > y (assumed case)
        JNE     NOT_LARGER_X    ; Go is x is not larger
        JMP     MOVE_2Z         ; Exit routine for x > y
NOT_LARGER_X:
        MOV     SI,CS:ENTRY_DI  ; y entry value to SI
        JMP     MOVE_2Z
;**********************|
;    offset smaller    |
;     significand      |
;**********************|
; DX holds the difference between exponents if this difference is 18
; or less. This value is used to offset the second significand
; SI --> BCD12 encoding of smaller number
NOT_TRIVIAL:
        LEA     DI,CS:SIG_S     ; Buffer for smaller number
        ADD     DI,DX           ; Add offset to destination
        ADD     SI,3            ; Mantissa digits in BCD12 encoding
        MOV     CX,9            ; 9 packed digits to move
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
;**********************|
;    test for unequal  |
;        signs         |
;**********************|
; The sign of the number is the high nibble of the first BCD12 digit
        MOV     SI,CS:ENTRY_SI  ; Set pointers to entry values
        MOV     DI,CS:ENTRY_DI
; SI --> x and DI --> y
        MOV     AL,BYTE PTR [SI]        ; x's first digit to AL
        MOV     AH,BYTE PTR [DI]        ; y's first digit to AH
        AND     AL,11110000B    ; Clear x's low nibble
        AND     AH,11110000B    ; and y's low nibble
        CMP     AL,AH           ; Test for equal sign
        JE      SAME_SIGN
        JMP     DIFFERENT_SIGN
;************************************|
; CASE 1: x and y have the same sign |
;************************************|
SAME_SIGN:
        MOV     CS:CASE_12,1    ; Case 1 processing switch
;**********************|
;   add significands   |
;**********************|
; At this point the significands of x and y are in the buffers SIG_L
; and SIG_S, ready for addition into SIG_R
        LEA     SI,CS:SIG_L     ; SI --> to larger significand
        LEA     DI,CS:SIG_S     ; DI --> to smaller significand
        LEA     BX,CS:SIG_R     ; BX --> buffer for sum
        ADD     SI,35           ; Point SI to last digit in buffer
        ADD     DI,35           ; also DI
        ADD     BX,36           ; and BX
        MOV     CX,36           ; 36 digits to add
        CLC                     ; Clear carry flag for first sum
ADD_36_DIGITS:
        MOV     AL,BYTE PTR CS:[SI]     ; One digit to AL
        ADC     AL,BYTE PTR CS:[DI]     ; Add other digit
        AAA                     ; Convert sum to unpacked BCD
        MOV     BYTE PTR CS:[BX],AL     ; Store digit in buffer
        DEC     SI              ; Bump pointers to next digit
        DEC     DI
        DEC     BX
        LOOP    ADD_36_DIGITS   ; Loop for 36 digits
; At this point carry is set if sum overflows 36 digits
        JNC     NO_OVF          ; Go if no overflow
        MOV     BYTE PTR CS:[BX],01H    ; Enter 1 if overlfow
        INC     CS:BIAS_Z       ; Adjust exponent
NO_OVF:
; SIG_R buffer now holds sum of significands
        JMP     FORM_BCD        ; Construct BCD12 result
;************************************|
; CASE 2: x and y have different sign|
;************************************|
DIFFERENT_SIGN:
        MOV     CS:CASE_12,2    ; Case 2 processing switch
;**********************|
;    discard x = y     |
;**********************|
; Signed addition of numbers with different signs and the same values
; result in zero
        CMP     CS:LARGER,'e'   ; Encoding for x = y
        JNE     X_NOTEQ_Y       ; Continue if not equal
; Clear result buffer to encode zero and exit
OUTPUT_ZERO:
        MOV     SI,CS:ENTRY_BX  ; Result buffer for BCD12 number
        MOV     CX,12           ; 12 bytes in BCD12 format
ZERO_RESULT:
        MOV     BYTE PTR [SI],00H       ; Place zero byte in buffer
        INC     SI              ; Bump buffer pointer
        LOOP    ZERO_RESULT     ; Loop until buffer cleared
        JMP     SIGN_ADD_EXIT
;**********************|
; subtract significands|
;**********************|
; At this point the significands of x and y are in the buffers SIG_L
; and SIG_S. The number with the larger exponent is in SIG_L and the
; number with the smaller exponent in SIG_S
X_NOTEQ_Y:
        LEA     SI,CS:SIG_L     ; SI --> to larger significand
        LEA     DI,CS:SIG_S     ; DI --> to smaller significand
        LEA     BX,CS:SIG_R     ; BX --> buffer for sum
        ADD     SI,35           ; Point SI to last digit in buffer
        ADD     DI,35           ; also DI
        ADD     BX,36           ; and BX
        MOV     CX,36           ; 36 digits to subtract
        CLC                     ; Clear carry flag for first sum
SUB_36_DIGITS:
        MOV     AL,BYTE PTR CS:[SI]     ; x digit to AL
        SBB     AL,BYTE PTR CS:[DI]     ; Subtract y digit
        AAS                     ; Convert difference to unpacked BCD
        MOV     BYTE PTR CS:[BX],AL     ; Store digit in buffer
        DEC     SI              ; Bump pointers to next digit
        DEC     DI
        DEC     BX
        LOOP    SUB_36_DIGITS   ; Loop for 36 digits
; At this point carry is set if there is a carry out of the last
; digit
        JNC     FORM_BCD        ; Go if no overflow
        MOV     BYTE PTR CS:[BX],01H    ; Enter 1 if overlfow
; SIG_R buffer now holds difference of significands
;*************************************|
;   Construct BCD12 number and exit   |
;*************************************|
FORM_BCD:
        CALL    ROUND_SIGF      ; Round off significand
        MOV     DI,CS:ENTRY_BX  ; Set pointer to result buffer
                                ; for BCD12 output
        ADD     DI,3            ; Point to significand area
;**********************|
;   store significand  |
;**********************|
; Significand of result is unpacked BCD form in the first 18 digits
; of the SIG_R buffer
; DI --> significand field of the BCD12 result buffer
        MOV     CX,9            ; Counter for 18 packed digits
        LEA     SI,CS:SIG_R     ; Pointer to unpacked digits
; Skip possible leftmost zero in buffer
        CMP     BYTE PTR CS:[SI],00H    ; Test for zero
        JNE     PACK_AND_MOVE   ; Go if no leading zero
        INC     SI              ; Skip zero
; Other leading zeros in buffer require adjustment of exponent
LEADING_ZEROS:
        CMP     BYTE PTR CS:[SI],00H    ; Test for leading zero
        JNE     PACK_AND_MOVE   ; Go if no leading zero
        INC     SI              ; Skip leading zero
        DEC     CS:BIAS_Z       ; Adjust exponent
        LOOP    LEADING_ZEROS
; At this point all significand is zeros, therefore the number is
; zero
        JMP     OUTPUT_ZERO     ; Zero result exit
PACK_AND_MOVE:
        MOV     AH,BYTE PTR CS:[SI]     ; Get one digit
        INC     SI              ; Bump pointer
        MOV     AL,BYTE PTR CS:[SI]     ; Get next digit
        INC     SI              ; Bump pointer to next digit
        PUSH    CX              ; Save counter
        MOV     CL,4            ; Count for 4 digits
        SHL     AH,CL           ; Shift high digit to high nibble
        POP     CX              ; Restore counter
        OR      AL,AH           ; Merge high nibble in AH with low
                                ; nibble in AH (pack digits)
        MOV     BYTE PTR [DI],AL        ; Store packed digits
        INC     DI
        LOOP    PACK_AND_MOVE
; All digits moved to the destination BCD12 buffer
        MOV     DI,CS:ENTRY_BX  ; Set pointer to result buffer
                                ; sign area
; Test case switch to determine sign of result
        CMP     CS:CASE_12,1    ; See if CASE 1
        JE      CASE1_SIGN      ; Go to sign routine for CASE 1
;**********************|
; CASE 2 sign routine  |
;**********************|
; In CASE 2 the sign of the result is the sign of the larger element
; The variable CS:LARGER holds 'x' if x > y
; Note that the possibily x = y is discarded previously
        CMP     CS:LARGER,'x'   ; Test for x > y
        JE      CASE1_SIGN      ; Use sign of x for result
; At this point the sign of y must be used for the result
; This sign is obtained from the entry value of DI
        MOV     SI,CS:ENTRY_DI  ; Set pointer the y element
        MOV     AL,BYTE PTR [SI]        ; Get sign byte
        AND     AL,11110000B            ; Isolate sign of number
        MOV     BYTE PTR [DI],AL        ; Store in result buffer
        JMP     EXPONENT_SIGN           ; Continue processing
;**********************|
; CASE 1 sign routine  |
;**********************|
; In CASE 1 both numbers have the same sign. The sign of the sum is
; the sign common to both elements
CASE1_SIGN:
        MOV     SI,CS:ENTRY_SI  ; Set pointer the x element
        MOV     AL,BYTE PTR [SI]        ; Get sign byte
        AND     AL,11110000B            ; Isolate sign of number
        MOV     BYTE PTR [DI],AL        ; Store in result buffer
;**********************|
; get sign of exponent |
;**********************|
EXPONENT_SIGN:
; Exponent of result is stored in BIAS_Z in bias 10000 form
; Values < 10000 indicate a negative exponent
        MOV     DX,CS:BIAS_Z    ; Exponent of result to DX
        CMP     DX,10000        ; Compare with bias
        JAE     GET_POS_VALUE   ; Go if positive exponent
; Positive exponent is already encoded in result buffer
; Encode negative exponent (0001B) into BCD12 number
; DI --> sign field
        MOV     AL,[DI]         ; Get sign digit into AL
        MOV     AH,00000001B    ; Code for negative exponent
        OR      AL,AH           ; OR exponent sign into byte
        MOV     BYTE PTR [DI],AL        ; Store in buffer
; Calculate absolute value of negative exponent
        MOV     AX,10000        ; Bias to AX
        SUB     AX,DX           ; Bias minus exponent
        XCHG    AX,DX           ; Move exponent to DX
        JMP     CALC_EXPONENT   ; Go to conversion routine
;**********************|
;  calculate value of  |
;  positive exponent   |
;**********************|
; DX = exponent in bias 10000 form
GET_POS_VALUE:
        MOV     AX,10000        ; Bias to AX
        SUB     DX,AX           ; Exponent minus bias
CALC_EXPONENT:
; DX = absolute value of exponent (in binary)
        MOV     AX,DX           ; For conversion routine
        PUSH    DI              ; Save pointer for result
        CALL    FAR PTR BIN_TO_BCD      ; Conversion routine
        POP     DI              ; Restore pointer
; DH and DL holds 4 packed BCD digits
        INC     DI              ; Bump pointer to exponent filed
        MOV     BYTE PTR [DI],DH        ; Store two digits
        INC     DI
        MOV     BYTE PTR [DI],DL        ; Store two other digits
        JMP     SIGN_ADD_EXIT
;
SIGN_ADD_BCD12  ENDP
;
;*******************************************************************
;**************************************|
; signed floating point BCD subtraction|
;**************************************|
SIGN_SUB_BCD12  PROC    FAR
; Subtraction of two signed BCD numbers stored in BCD12 floating
; point format
; Operation:
;           z = x - y
;           where x, y, and z are signed, floating point numbers
; On entry:
;          DS:SI --> Minuend (element x in z = x - y)
;          DS:DI --> Subtraend (element y in z = x - y)
;          DS:BX --> 12-byte storage area for difference (element z)
; Note: the code assumes that the BCD12 numbers are in normalized
;       form, that is, that there are no leading zeros in the
;       significand
; On exit:
;       DS:BX --> difference (element z in z = x - y)
;       Entry pointers are preserved
; Operation:
;   Processing is based on the algebraic priciple of changing the
;   sign of the subtraend and proceeding as in addition
;
; change subtraend sign
        CALL    REVERSE_SIGN    ; Procedure to reverse the sign of
                                ; a BCD12 number
; Call signed addition routine
        CALL    SIGN_ADD_BCD12
; Restore sign of subtraend to entry value
        CALL    REVERSE_SIGN
        RET
SIGN_SUB_BCD12  ENDP
;*******************************************************************
;                     auxiliary procedures
;*******************************************************************
STORE_AND_CLEAR         PROC    NEAR
; Store entry pointers in scratchpad area and clear work buffers
;**********************|
; store entry pointers |
;**********************|
        MOV     CS:ENTRY_SI,SI          ; Store SI pointer
        MOV     CS:ENTRY_DI,DI          ; Store DI pointer
        MOV     CS:ENTRY_BX,BX          ; Store BX pointer
;**********************|
;  clear work buffers  |
;**********************|
        PUSH    SI              ; Save x value entro pointer
        LEA     SI,CS:SIG_L     ; Pointer to buffer
        MOV     CX,109          ; Clear three buffers
CLEAR_109:
        MOV     BYTE PTR CS:[SI],00H    ; Clear one byte in buffer
        INC     SI                      ; Bump pointer
        LOOP    CLEAR_109
        POP     SI              ; Restore x pointer
        RET
STORE_AND_CLEAR         ENDP
;*******************************************************************
;
REVERSE_SIGN    PROC    NEAR
; Invert sign of a number encoded in BCD12 format
; On entry:
;         DI --> number
; On exit: 
;         DI --> number with sign reversed
;
        TEST    BYTE PTR [DI],00010000B ; Bit 4 set if sign is -
        JNZ     SIGN_MINUS      ; Go if bit set
;**********************|
;    change + to -     |
;**********************|
        OR      BYTE PTR [DI],00010000B ; Make negative
        RET
;**********************|
;    change - to +     |
;**********************|
SIGN_MINUS:
        AND     BYTE PTR [DI],00001111B ; Make postive
        RET
REVERSE_SIGN    ENDP
;*******************************************************************
;
ROUND_SIGF      PROC    NEAR
; Rounding of significand digits stored in SIG_R buffer
; Rationale:
; SIG_R holds 37 digits in unpacked BCD format. The BCD12 format
; encodes 18 siginificand digits
; Rounding proceeds with the 19th digit in SIG_R, as follows:
; 1. If the 19th digit is < 5 then no rounding is required
; 2. If the 19th digit is > 5 then the first 18 digits are rounded
; 3. If the 19th digit = 5 then the digits to the right are examined.
;    If the digits to the right are all zeros (5000..00) then the
;    18th digit is incremented only if it is odd. This rounding
;    method is known as "round to nearest even".
; 4. If the result of incrementing the 18th digit is 10, the digit
;    to the left is also incremented, until no carry is generated.
;
        LEA     SI,CS:SIG_R     ; Set pointer to buffer in CS
        ADD     SI,18           ; Index to the 19th digit
        MOV     AL,CS:[SI]      ; 19th digit to AL
        CMP     AL,05H          ; Test for 5 or greater
        JB      EXIT_ROUNDING   ; No rounding necessary
; 19th digit is 5 or greater
        JE      MIDWAY_CASE     ; Go to "round to even" routine
; At this point the 19th digit is > 5
;**********************|
;  bump digit and test |
;**********************|
ROUND_OFF:
        DEC     SI              ; Point to 18th digit
ROUND_EVEN:
        MOV     AL,CS:[SI]      ; Get BCD digit into AL
        ADD     AL,01H          ; Add 1 to digit
        AAA                     ; Adjust result
        JNC     EXIT_ROUNDING   ; Go if no carry generated
        JMP     ROUND_OFF
; Note that no testing for the end of buffer is required since the
; leftmost digit in SIG_R cannot be greater than 8
;**********************|
; test for midway case |
;**********************|
; SI --> to the 19th digit
MIDWAY_CASE:
        PUSH    SI              ; Save pointer to 19th digit
        MOV     CX,17           ; Set counter for 17 digits
TEST_FOR_0H:
        INC     SI              ; Bump pointer to next digit
        CMP     BYTE PTR CS:[SI],00H    ; Test for a zero digit
        JNE     NOT_MIDWAY      ; Exit if not midway case
        LOOP    TEST_FOR_0H     ; Continue testing
; At this point the undisplayed digits are in the pattern 5000..00
        POP     SI              ; Restore pointer to 19th digit
        DEC     SI              ; Point to 18th digit
        MOV     AL,CS:[SI]      ; Get 18th digit into AL
;**********************|
;  test for odd 18th   |
;        digit         |
;**********************|
; Binary encodings of odd numbers have the LOB set
        TEST    AL,00000001B    ; Test LOB to determine if odd
        JNZ     ROUND_EVEN      ; Go to rounding routine if odd
        JMP     EXIT_ROUNDING   ; Do not round if already even
;**********************|
;    not midway case   |
;**********************|
; The number is rounded if the undiplayed significand is not in the
; form 50000..00
NOT_MIDWAY:
        POP     SI              ; Restore pointer
        JMP     ROUND_OFF       ; Go to rounding routine
;**********************|
;   exit procedure     |
;**********************|
EXIT_ROUNDING:
        RET
ROUND_SIGF      ENDP
;*******************************************************************
;
UNPACK_MOVE     PROC    NEAR
; Move the number of packed digits in CX from DS:SI to CS:DI 
; and unpack
; On entry:
;         CX = number of packed BCD digits to be moved
;         DS:SI --> packed BCD string to move
;         CS:DI --> destination of unpacked string
; On exit:
;         CS:DI = unpacked string of BCD digits
;
        PUSH    SI
        PUSH    DI
; Move CX digits
MOVE_DIGS:
        MOV     AL,[SI]         ; Get one byte from source
; AL has two packed BCD digits
        MOV     AH,AL           ; Copy packed digits to AH
        AND     AL,00001111B    ; Clear high nibble to isolate low
                                ; BCD digit
        PUSH    CX              ; Save digits counter
        MOV     CL,4            ; Counter for 4 bits
        SHR     AH,CL           ; Shift right to isolate high BCD
                                ; digit
        POP     CX              ; Restore counter
; AH holds the unpacked high digit and AL holds the low digit
        MOV     CS:[DI],AH      ; High BCD digit to buffer
        INC     DI              ; Bump buffer pointer
        MOV     CS:[DI],AL      ; Low BCD digit to buffer
        INC     DI              ; Bump pointer again
        INC     SI              ; Bump source pointer
        LOOP    MOVE_DIGS
        POP     DI
        POP     SI
        RET
UNPACK_MOVE     ENDP
;*******************************************************************
;
TEST_4_ZERO     PROC    NEAR
; Test a number encoded in BCD12 format for a zero value
; Note: a number in BCD12 format is zero if all numbers in the
;       exponent and significand are zero
; On entry:
;         SI --> Number stored in BCD12 format
; On exit:
;         Carry set if number is zero
;         Carry clear if number is not zero
;
        PUSH    SI              ; Save entry pointer
        INC     SI              ; Point to first exponent digit
        MOV     CX,11           ; Counter for 11 bytes to test
TEST_11_ZERO:
        CMP     BYTE PTR[SI],0  ; Is the byte zero
        JNE     NOT_11_ZERO     ; Go if not zero
        INC     SI              ; Bump pointer
        LOOP    TEST_11_ZERO    ; Test all digits
; At this point the number is zero
        POP     SI              ; Restore entry pointer
        STC                     ; Set zero return signal
        RET
NOT_11_ZERO:
        POP     SI              ; Restore entry pointer
        CLC                     ; Set non-zero return signal
        RET
TEST_4_ZERO     ENDP
;*******************************************************************
;
EXP_2_BIAS      PROC    NEAR
; Convert exponent of BCD12 number into binary and add bias
; On entry:
;         SI --> Number encoded in BCD12 format
; On exit:
;         DX = binary value of exponent
;
        PUSH    BX              ; Save pointers
        PUSH    SI
        PUSH    DI
        INC     SI              ; Pointer to exponent field
        INC     SI              ; Second exponent digit
; Move BCD digits to AL-BH-BL-CH-CL for conversion routine
        MOV     AL,0            ; Highest digit is zero
        MOV     DL,BYTE PTR [SI]        ; DL holds 2 packed digits
        MOV     DH,DL           ; Copy digits to DH
        AND     DL,00001111B    ; Erase high nibble in DL
        AND     DH,11110000B    ; and low in DH
        MOV     CX,4            ; For a count of 4
        SHR     DH,CL           ; Shift digit to low position
        DEC     SI              ; Bump pointer to previous digit
        MOV     BL,BYTE PTR [SI]        ; BL holds 2 packed digits
        MOV     BH,BL           ; Copy digits to BH
        AND     BL,00001111B    ; Erase high nibble in BL
        AND     BH,11110000B    ; and low in BH
        MOV     CX,4            ; For a count of 4
        SHR     BH,CL           ; Shift digit to low position
        MOV     CX,DX           ; Move digits to CX
        CALL    FAR PTR BCD_INTO_REG    ; BCD to binary conversion
        POP     DI              ; Restore entry pointers
        POP     SI
        POP     BX
; At this point DX hold binary value for exponent
; SI --> sign byte
; Test for negative exponent. Add bias if positive
        TEST    BYTE PTR[SI],01H        ; Test for negative exponent
        JNZ     EXP_IS_NEG
; At this point exponent is positive
        ADD     DX,10000        ; Add bias to exponent value
        RET                     ; End of processing
; Subtract exponent from bias
EXP_IS_NEG:
        MOV     AX,10000        ; Bias
        SUB     AX,DX           ; Value minus bias
        MOV     DX,AX           ; Result to DX
        RET                     ; End of processing
EXP_2_BIAS      ENDP
;*****************************************************************
;
ID_CPU          PROC    FAR
; Procedure to determine the CPU in an IBM microcomputers
; On exit:
;           AL = 1 if CPU is 8086 or 8088
;           AL = 2 if CPU is 80286
;           AL = 3 if CPU is 80386
;           AL = 4 if CPU is 486
;           AL = 5 if CPU is Pentium
;
;***********************|
;   test for 8086/8088  |
;***********************|
; Bits 12 to 15 in the flag register are always set in the 8086
; and 8088 CPU
        PUSHF                   ; Flag register to stack
        POP     AX              ; Store flags in AX
        AND     AX,0FFFH        ; Clear bits 12 to 15
        PUSH    AX              ; AX to stack
        POPF                    ; and to flags register
        PUSHF                   ; Flags to stack
        POP     AX              ; and to AX for reading
        AND     AX,0F000H       ; Preserve bits 12 to 15
        CMP     AX,0F000H       ; Test for bits set
        JNE     TEST_286        ; Go if bits not set
; At this point processor is a 8086 or 8088
        MOV     AX,1            ; Return code
        JMP     ID_EXIT         ; Exit
;***********************|
;     test for 80286    |
;***********************|
; Bits 12 to 15 in the flag register are always clear in the Intel
; 80286 CPU
TEST_286:
        PUSHF                   ; Flag register to stack
        POP     BX              ; Store flags in BX
        OR      BX,0F000H       ; Make sure bit field is set
        PUSH    BX              ; To stack
        POPF                    ; And to flag register
        PUSHF                   ; Flags to stack
        POP     AX              ; And to AX
        AND     AX,0F000H       ; Clear all other bits
        JNZ     TEST_386        ; Go if bits not clear
; At this point processor is an 80286
        MOV     AX,2            ; Return code
        JMP     ID_EXIT         ; Exit
;***********************|
;     test for 80386    |
;***********************|
.386
; Note: The .386 assembler directive is compatible with the MASM
;       assembler version 5.0 and later. The directive is
;       inserted at this point in the listing so that the
;       preceding code will execute in 8086/8088 and 80286
;       machines.
;
; Bit 18 of the E flags register was introduced in the 486 CPU
; This bit cannot be set in the 80386
TEST_386:
        PUSHFD                  ; 32-bits E flags to stack
        POP     EAX             ; Flags to EAX
        OR      EAX,40000H      ; Make sure bit 18 is set
        PUSH    EAX             ; New flags to stack
        POPFD                   ; An to E flags register
        PUSHFD                  ; Back to stack
        POP     EAX             ; And to EAX
        AND     EAX,40000H      ; Clear all except bit 18
        JNZ     TEST_486        ; Go if bit 18 is clear
; At this point processor is a 80386
        MOV     AX,3            ; Return code
        JMP     ID_EXIT         ; Exit
;***********************|
;     test for 486      |
;***********************|
; Bit 21 (ID flag) of the E flags register cannot be set in the
; 486
TEST_486:
        PUSHFD                  ; 32-bits E flags to stack
        POP     EAX             ; Flags to EAX
        OR      EAX,200000H     ; Make sure bit 21 is set
        PUSH    EAX             ; New flags to stack
        POPFD                   ; An to E flags register
        PUSHFD                  ; Back to stack
        POP     EAX             ; And to EAX
        AND     EAX,200000H     ; Clear all except bit 21
        JNZ     IS_PENTIUM      ; Go if bit 21 is clear
; At this point processor is a 486
        MOV     AX,4            ; Return code
        JMP     ID_EXIT         ; Exit
;***********************|
;  processor is PENTIUM |
;***********************|
IS_PENTIUM:
        MOV     AX,5            ; Return code
ID_EXIT:
        RET
ID_CPU          ENDP
.8086
;*****************************************************************
;
; ID_FPU  PROCEDURE
; Determine if there is a mathematical coprocessor installed and
; if it is an 8087, 80287, 80387, or the math unit of a 486 or
; Pentium
; On exit:
;         AL = 0 if no coprocessor present
;              1 if 8087
;              2 if 80287
;              3 if 80387
;              4 if math unit of 486 or 486 SX/387 system
;              5 if math unit of Pentium
;****************************************************************
; Code segment storage for 80x87 control and status words
CONTROL_87      DW      0 
STATUS_87       DW      0
;****************************************************************
;
ID_FPU          PROC    FAR
        FNINIT          ; Initialize coprocessor (if present)
; Note that the no-wait form must be used to prevent a wait
; forever condition
        MOV     AX,5A5AH    ; Value to set in status word
        MOV     WORD PTR CS:STATUS_87,AX
        FNSTSW  CS:STATUS_87    ; Store status word
        MOV     AX,CS:STATUS_87 ; Read status into AX
        CMP     AL,0            ; Test for no status bits
        JE      CHK_CONTROL     ; Go if 0
; At this point to FPU unit is detected in syste
        MOV     AL,0            ; Return code for no coprocessor
        RET
; A secondary test is based on the FPU control word
CHK_CONTROL:
        FNSTCW  CS:CONTROL_87   ; Store control word
        MOV     AX,CS:CONTROL_87        ; Read into AX
        CMP     AL,0            ; Test for no status bits
        AND     AX,103FH        ; Bit mask is
                                ; AH = 0001 0000
                                ; AL = 0011 1111
        CMP     AX,003FH        ; Test for AL bits unchanged
        JE      SYS_80X87       ; 80x87 is present
; At this point no FPU unit is detected in system
        MOV     AL,0            ; Return code for no coprocessor
        RET
;**********************|
;  coprocessor present |
;**********************|
; The first test for the coprocessor type can be based on the CPU
; type installed in the host
        CALL    ID_CPU          ; Library procedure to identify
                                ; Intel CPU
; At this point AL holds CPU ID code as follows:
;           AL = 1 if FPU is 8087
;           AL = 2 if FPU is 80287
;           AL = 3 if FPU is 80287 or 80387 since the 80287 can
;                be installed in an 80386 system
;           AL = 4 if FPU is 486 math unit
;           AL = 5 if FPU is Pentium math unit
;
; If return code is not 3 then there is no uncertainty regarding
; the FPU type
SYS_80X87:
        CMP     AL,3            ; Test for undefined case
        JNE     IS_386          ; Go if 386 CPU
; If not 386, return with coprocessor code in AL
        RET
;**********************|
;    80287 or 80387    |
;        system        |
;**********************|
; The test for an 80287 system is based on the state of the
; condition code bits after FINIT. In the 80387 C0 to C3 bits
; are zero after FINIT, in the 80287 they contain the previous
; value
; Note: at this time the wait form of FPU opcodes can be used
;       since the presence of a coprocessor is certain
IS_386:
        FINIT           ; Make sure stack is EMPTY
        FXAM            ; C0 and C3 are set after FXAM if
                        ; the stack top is EMPTY
        FINIT           ; In an 80287 C0 and C3 will remain
                        ; set after the second FINIT
        FSTSW   CS:STATUS_87    ; Status word to memory
        FWAIT                   ; Synchronize CPU and 80x87
        MOV     AX,CS:STATUS_87 ; Status word to AX register
        AND     AH,01000001B    ; Clear all bits except C3
                                ; and C0
        JNZ     TYPE_287        ; Go if bits not clear
;**********************|
;     80387 or 486     |
;**********************|
        MOV     AL,3            ; Return code for 80387
        RET
;**********************|
;        80287         |
;**********************|
TYPE_287:
        MOV     AL,2            ; Return code for 80287
        RET
ID_FPU  ENDP
;*****************************************************************
;
;NUM_AT_ST0 PROCEDURE
; Procedure to identify and display the number type at ST(0) using
; a jump based on the condition code settings
;
; On entry:
;         DS:DI -> caller's text buffer with a minimum of 12
;                  free bytes for message storage
; On exit:
;                                         text message (DS:DI)
;   AL = 0 for positive normal          + normal
;        1 for negative normal          - normal
;        2 for positive denormal        + denormal
;        3 for negative denormal        - denormal
;        4 for positive zero            + zero
;        5 for negative zero            - zero
;        6 for positive NAN             + NAN
;        7 for negative NAN             - NAN
;        8 for positive infinity        + infinity
;        9 for negative infinity        - infinity
;       10 for empty                    EMPTY
;       11 for unsupported type         unsupported
;
;      AX is destroyed. All other registers are preserved
;
;
;***************************|
;    code segment data      |
;***************************|
; Jump table for processing routines
;                              |----> LABELS IN CODE SEGMENT
;                              |
;                       |--------------|; CC bits = value
EXAM_TBL        DD      UNSUPPORTED     ; 0 0 0 0 = 0
                DD      POS_NAN         ; 0 0 0 1 = 1
                DD      UNSUPPORTED     ; 0 0 1 0 = 2
                DD      NEG_NAN         ; 0 0 1 1 = 3 
                DD      POS_NORMAL      ; 0 1 0 0 = 4
                DD      POS_INFINITY    ; 0 1 0 1 = 5
                DD      NEG_NORMAL      ; 0 1 1 0 = 6
                DD      NEG_INFINITY    ; 0 1 1 1 = 7
                DD      POS_ZERO        ; 1 0 0 0 = 8
                DD      EMPTY           ; 1 0 0 1 = 9
                DD      NEG_ZERO        ; 1 0 1 0 = 10
                DD      EMPTY           ; 1 0 1 1 = 11
                DD      POS_DENORMAL    ; 1 1 0 0 = 12
                DD      EMPTY           ; 1 1 0 1 = 13 
                DD      NEG_DENORMAL    ; 1 1 1 0 = 14
                DD      EMPTY           ; 1 1 1 1 = 15
;
; Error message table. Each entry is 12 bytes
ERR_MESS_TBL    DB      '+ normal    '
                DB      '- normal    '
                DB      '+ denormal  '
                DB      '- denormal  '
                DB      '+ zero      '
                DB      '- zero      '
                DB      '+ NAN       '
                DB      '- NAN       '
                DB      '+ infinity  '
                DB      '- infinity  '
                DB      'EMPTY       '
                DB      'unsupported '
;
; Storage for FPU status word
FPU_STATUS      DW      0       ; Storage for status word
;*****************************************************************
;
NUM_AT_ST0      PROC    FAR
;***************************|
;   save caller's machine   |
;         state             |
;***************************|
        PUSH    BX              ; Caller's environment
        PUSH    CX
        PUSH    DX
        PUSH    SI
        PUSH    DI
;***************************|
;     get condition codes   |
;***************************|
; Get condition codes for stack top register and store for later
; reference
        FXAM                    ; Examine condition codes
        FWAIT                   ; For 8087 compatibility
        FSTSW   CS:FPU_STATUS   ; Store status word in memory
        FWAIT
;***************************|
;   store condition codes   |
;          in BX            |
;***************************|
;  AH bits: 7  6  5  4  3  2  1  0
;              3           2  1  0  condition code bits
;              |___________|__|__|_____ Condition code settings:
        MOV     BX,CS:FPU_STATUS        ; Status word to BX
; Clear all non-condition code bits
; Note: shifts are coded with the operand in CL so that the code
;       will be compatible with 8086/8088//80286 systems
        AND     BH,01000111B            ; Mask to clear bits
        MOV     CL,6            ; Shift count
        SHR     BX,CL           ; C0-C2 now 000x xx00 in BL
        MOV     CL,5            ; Shift count
        SAL     BH,CL           ; C3 now 00x0 0000 in BH
        OR      BL,BH           ; C0-C3 now 00xx xx00 in BL
        XOR     BH,BH           ; Clear high byte
;***************************|
;   jump based on CC bits   |
;***************************|
        JMP     CS:EXAM_TBL [BX]   ; Jump using CC bits
;************************************|
;    destination labels for jumps    |
;************************************|
; Each jump label sets the corresponding error code in AX and
; exits to the epilogue routine
UNSUPPORTED:
        MOV     AX,11
        JMP     EXIT_RTN
POS_NAN:
        MOV     AX,6
        JMP     EXIT_RTN
NEG_NAN:
        MOV     AX,7
        JMP     EXIT_RTN
POS_NORMAL:
        MOV     AX,0
        JMP     EXIT_RTN
POS_INFINITY:
        MOV     AX,8
        JMP     EXIT_RTN
NEG_NORMAL:
        MOV     AX,1
        JMP     EXIT_RTN
NEG_INFINITY:
        MOV     AX,9
        JMP     EXIT_RTN
POS_ZERO:
        MOV     AX,4
        JMP     EXIT_RTN
EMPTY:
        MOV     AX,10
        JMP     EXIT_RTN
NEG_ZERO:
        MOV     AX,5
        JMP     EXIT_RTN
POS_DENORMAL:
        MOV     AX,2
        JMP     EXIT_RTN
NEG_DENORMAL:
        MOV     AX,3
;***************************|
;  move message to caller's |
;     buffer and exit       |
;***************************|
; At this point AX holds the error code number
; The value in AX is multiplied by 12 to obtain the offset
; of the corresponding error message
; Message is then moved to the caller's buffer by DS:DI
EXIT_RTN:
        PUSH    AX              ; Save error code
        LEA     SI,CS:ERR_MESS_TBL      ; Offset of table
        MOV     CL,12           ; Length of each message
        MUL     CL              ; AX = offset of correct message
        ADD     SI,AX           ; Add to table offset
; At this point:
;       CS:SI --> 12 byte number type message
;       DS:DI --> caller's buffer with 12 bytes minimum space
        MOV     CX,12           ; Counter for 12 bytes
TRANSFER_12:
        MOV     AL,CS:[SI]      ; Get message character
        MOV     DS:[DI],AL      ; Place in caller's buffer
        INC     SI              ; Bump buffer pointers
        INC     DI
        LOOP    TRANSFER_12
        POP     AX              ; Restore error code register
        POP     DI              ; Restore caller's environment
        POP     SI
        POP     DX
        POP     CX
        POP     BX
        RET
NUM_AT_ST0      ENDP
;
;
CODE    ENDS
        END

     POP     DX
        POP     CX
        POP     BX
        RET
NUM_AT_ST0      ENDP
;
;
CO