;*****************************************************************
;*****************************************************************
;                           UNIT_5.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  March 4, 1994
; Updated:        March 6, 1994
;
; Libary name: MATHUNIT.LIB
;
;*****************************************************************
; MATHEMATICAL ROUTINES IN THIS MODULE
PUBLIC TEN_TO_Y_BYFAC   ; .. 10**y by exponent factoring
PUBLIC X_TO_Y_BYFAC     ; .. x**y by exponent factoring
PUBLIC X_TO_Y_BYBP      ; .. x**y by binary powering
PUBLIC X_TO_Y_BYLOG     ; .. x**y by logarithms
PUBLIC X_TO_Y_MIXED     ; .. x**y by logs and binary powering
PUBLIC LOG_10           ; .. base 10 logarithms
PUBLIC LOG_E            ; .. base e logarithms
PUBLIC ALOG_10          ; .. base 10 antilogarithm
PUBLIC ALOG_E           ; .. base e antilogarithm
PUBLIC ALOG_2           ; .. base 2 anitlogarithm
PUBLIC INPUT_NUMBER     ; .. super procedure for input
PUBLIC OUTPUT_NUMBER    ; .. super procedure for output
;
;*****************************************************************
;                          externals
;*****************************************************************
; Routines in other modules of MATHUNIT.LIB
EXTRN   FPU_INPUT:      FAR
EXTRN   FPU_OUTPUT:     FAR
EXTRN   ASCII_TO_EXP:   FAR
;
;****************************************************************
;                          code segment
;****************************************************************
;
CODE    SEGMENT BYTE
        ASSUME  CS:CODE
;
;****************************************************************
;                  10 to the y by exponent factoring
;****************************************************************
;***************************|
;    code segment constants |
;***************************|
; Tables of constants holding 40 values, 10 for each processing
; stage
TABLE_1000      DT      0, 1.0E1000, 1.0E2000, 1.0E3000, 1.0E4000
TABLE_100       DT      0, 1.0E100, 1.0E200, 1.0E300, 1.0E400
                DT      1.0E500, 1.0E600, 1.0E700, 1.0E800
                DT      1.0E900
TABLE_10        DT      0, 1.0E10, 1.0E20, 1.0E30, 1.0E40
                DT      1.0E50, 1.0E60, 1.0E70, 1.0E80
                DT      1.0E90
TABLE_1         DT      0, 1.0E1, 1.0E2, 1.0E3, 1.0E4
                DT      1.0E5, 1.0E6, 1.0E7, 1.0E8, 1.0E9
;
Y_VALUE         DW      0
;
TEN_TO_Y_BYFAC   PROC   FAR
; Exact calculation of 10**y by exponent factorization according
; to the following table
;
; F3 = 1000    I3 = INT(y/F3)            R3 = y MOD(I3*F3)
; F2 = 100     I2 = INT(R3/F2)           R2 = R3 MOD(I2*F2)
; F1 = 10      I1 = INT(R2/F1)           R1 = R2 MOD(I1*F1)
; F0 = 1       I0 = INT(R1/F0)

; Exponent factorization formula:
;
; 10**y = (10**F3)**I3 * (10**F2)**I2 * (10**F1)**I1 * (10**F0)**I0
;         ____________   ____________    ___________    ___________
;               |              |               |              |
;               |              |               |              |
;            STAGE 3        STAGE 2         STAGE 1       STAGE 0
; partial       |              |               |              |
; products     S3             S2              S1             S0 
;           ------------------------------------------------------
;                      e x e c u t i o n    s t a g e s
;
; In this variant of the exponent factoring algorithm all partial
; products are obtained from RAM tables
;
; On entry:
;         ST(0) = exponent (y = word integer)
;
; On exit:
;         ST(0) = 10**y
;
;*****************************|
; move data to work variable  |
;*****************************|
        FISTP   CS:Y_VALUE      ; Storage for y (integer)
        MOV     CX,CS:Y_VALUE       ; Exponent to loop register
;***************************|
;        special case       |
;       test for x**0       |
;***************************|
; By the rules of exponents any base to the 0 power = 1
        CMP     CX,0            ; Test for 0 exponent
        JNE     EXP_OTHER       ; Go if not zero
; At this point we have detected x**0
        FLD1                    ; Load 1
        RET
;*****************|
;     STAGE 3     |
;********************************|
;    calculate (10**F3)**I3      |
;********************************|
; Test for y > 1000
; First load 1 for the case that exp < 1000
EXP_OTHER:
        FLD1                    ;    1    | ------- | ------- |
        CMP     CX,1000         ; Is exp > 1000
        JNA     STAGE_2V3       ; Go if not above than 100
        FSTP    ST(0)           ; ------- | ------- | ------- |
; At this point CX is > 1000
;**************************|
;   calculate element I3   |
;      INT (y/1000)        |
;**************************|
; The division by 1000 is performed in 2 steps in order to avoid
; a divide overflow
        MOV     DX,CX           ; Save original exponent in DX
        MOV     AX,CX           ; Exponent to AX
        MOV     BX,100          ; First divisor
        DIV     BL              ; Exponent / 100
        MOV     AH,0            ; Clear remainder of byte division
        MOV     BX,10           ; Second divisor
        DIV     BL              ; First dividend / 10
        MOV     AH,0            ; Clear remainder of byte division
        MOV     CX,AX           ; Copy divisor to CX
; Calculate R3 = y MOD (I3 * F3)
; At this point:
;               DX = y
;               CX = INT(y/1000) = I3
        PUSH    DX              ; Save y in stack
        MOV     AX,1000         ; F3 to AX
        MUL     CX              ; AX = I3 * F3
        POP     DX              ; Restore y
        SUB     DX,AX           ; DX = y MOD (I3 * F3)
; CX = INT(y/1000)
        MOV     AX,CX           ; AX = y MOD 1000 (not 0)
        MOV     CL,10           ; Multiplier to CL
        MUL     CL              ; AX x 10 = offset in table
        MOV     SI,AX           ; Offset to index register
        FLD     CS:TABLE_1000[SI]
; Test for exponent exact multiple of 1000 (no remainder)
        MOV     CX,DX           ; R3 to CX
        CMP     CX,0            ; In this case CX = 0
        JNE     STAGE_2V3       ; Go if not
;*************************|
;  quick exit if REM = 0  |
;***************************************************|
;   exponent factor as follows:                     |
;      (10**F3)**I3 ------> S3 ---------->   ST(0)  |
;***************************************************|
; Return if exponent is exact multiple of 1000
        RET                     ; 10**y   | ------- |
;
;*****************|
;     STAGE 2     |
;********************************|
;    calculate (10**F2)**I2      |
;********************************|
; At this point:                |  ST(0)  |  ST(1)  | ST(2)  |
;                               |   S3    | ------- |
;                               |  or 1   |
STAGE_2V3:
; Test for EXP > 100
; First load 1 for the case that exp < 100
        FLD1                    ;    1    |   S3    | ------- |
        CMP     CX,100          ; Is exp > 100
        JNA     STAGE_1V3       ; Go if not above than 100
        FSTP    ST(0)           ;    S3   | ------- | ------- |
; At this point CX is > 100
;**************************|
;   calculate element I2   |
;      INT (y/100)         |
;**************************|
; CX = INT(y/1000)
        MOV     AX,CX           ; Move S3 to AX
        MOV     DX,CX           ; Update DX
        MOV     BX,100          ; Divisor
        DIV     BL              ; Exponent / 100
        MOV     AH,0            ; Clear reminder
        MOV     CX,AX           ; Copy divisor to CX
; Calculate R2 = R3 MOD (I2 * F2)
; At this point:
;               DX = R3
;               CX = INT(y/100) = I2
        PUSH    DX              ; Save y in stack
        MOV     AX,100          ; F2 to AX
        MUL     CL              ; AX = I2 * F2
        POP     DX              ; Restore y
        SUB     DX,AX           ; DX = R3 MOD (I2 * F2)
; CX = INT(y/100)
        MOV     AX,CX           ; AX = REM y MOD 25 (not 0)
        MOV     CL,10           ; Multiplier to CL
        MUL     CL              ; AX x 10 = offset in table
        MOV     SI,AX           ; Offset to index register
        FLD     CS:TABLE_100[SI]
; AX holds R2
        MOV     CX,DX           ; Copy S2 to CX
; Test for exponent exact multiple of 100 (no remainder)
        CMP     CX,0            ; In this case CX = 0
        JNE     STAGE_1V3       ; Go if not
;*************************|
;  quick exit if REM = 0  |
;***************************************************|
;   multiply exponent factors:                      |
;    I3(10**F3) --------> S3 ------------>   ST(1)  |
;    I3(10**F2) --------> S2 ------------>   ST(0)  |
;***************************************************|
; Return if exponent is exact multiple of 100
;                               |    S3    |    S2   | -------- |
        FMUL    ST,ST(1)
        FSTP    ST(1)           ; 10**y   | ------- | -------- |
        RET                     ; Exponent is exact multiple of 100
;
;*****************|
;     STAGE 1     |
;********************************|
;    calculate (10**F1)**I1      |
;********************************|
; At this point:                |  ST(0)  |  ST(1)  |  ST(2)  |
;                               |   S2    |   S3    | ------- |
;                               |  or 1   |   or 1
; CX = INT(y/10)
; Test for EXP > 10
; First load 1 for the case that exp < 10 
STAGE_1V3:
        FLD1                    ;    1    |    S2   |    S3   |
        CMP     CX,10           ; Is exp > 25 
        JNA     STAGE_0V3       ; Go if not above 10
        FSTP    ST(0)           ;   S2    |    S3   | ------- |
; At this point CX is > 10
;**************************|
;   calculate element I1   |
;      INT (y/10)          |
;**************************|
; CX = INT(y/100)
        MOV     AX,CX           ; New exponent to AX
        MOV     DX,CX           ; Update DX
        MOV     BX,10           ; Divisor
        DIV     BL              ; Exponent / 25 
        MOV     AH,0            ; Clear reminder
        MOV     CX,AX           ; Copy divisor to CX
; Calculate R1 = R2 MOD (I1 * F1)
; At this point:
;               DX = R2
;               CX = INT(y/10) = I1
        PUSH    DX              ; Save y in stack
        MOV     AX,10           ; F1 to AX
        MUL     CL              ; AX = I1 * F1
        POP     DX              ; Restore y
        SUB     DX,AX           ; DX = R2 MOD (I1 * F1)
; CX = INT(y/10)
        MOV     AX,CX           ; AX = REM y MOD 25 (not 0)
        MOV     CL,10           ; Multiplier to CL
        MUL     CL              ; AX x 10 = offset in table
        MOV     SI,AX           ; Offset to index register
        FLD     CS:TABLE_10[SI]
; DX holds remainder
        MOV     CX,DX           ; R1 to CX
; Test for exponent exact multiple of 10
        CMP     CX,0            ; In this case CX = 0
        JNE     STAGE_0V3       ; Go if not
;*************************|
;  quick exit if REM = 0  |
;***************************************************|
;   multiply exponent factors:                      |
;    I3(10**F3) --------> S3 ------------>   ST(2)  |
;    I2(10**F2) --------> S2 ------------>   ST(1)  |
;    I1(10**F1) --------> S1 ------------>   ST(0)  |
;***************************************************|
; Return if exponent is exact multiple of 10
;                               |    S1   |    S2   |    S3   |
        FMUL    ST,ST(1)
        FSTP    ST(1)           ;    PP   |    S3   | ------- |
        FMUL    ST,ST(1)
        FSTP    ST(1)           ; 10**y   | ------- | -------- |
        RET
;
;*****************|
;     STAGE 0     |
;********************************|
;    calculate (10**F0)**I0      |
;********************************|
; At this point:                |  ST(0)  |  ST(1)  |  ST(2)  |
;                               |   S1    |    S2   |   S3    |
;                               |  or 1   |   or 1  |   or 1  |
; CX = INT(y/10)
STAGE_0V3:
        MOV     AX,CX           ; AX = REM y MOD 25 (not 0)
        MOV     CL,10           ; Multiplier to CL
        MUL     CL              ; AX x 10 = offset in table
        MOV     SI,AX           ; Offset to index register
        FLD     CS:TABLE_1[SI]
;***************************************************|
;   multiply exponent factors:                      |
;      (10**F3)**I3 -------> P3 --------->   ST(3)  |
;      (10**F2)**I2 -------> P2 --------->   ST(2)  |
;      (10**F1)**I1 -------> P1 --------->   ST(1)  |
;      (10**F0)**I0 -------> P0 --------->   ST(0)  |
;***************************************************|
;                               |    P0    |    P1    |    P2    |.. 
        FMUL    ST,ST(1)
        FSTP    ST(1)           ;    PP    |    P2    |    P3    |
        FMUL    ST,ST(1)
        FSTP    ST(1)           ;    PP    |    P3    | -------- |
        FMUL    ST,ST(1)
        FSTP    ST(1)           ;  10**y   | -------- | -------- |
        RET
;
TEN_TO_Y_BYFAC   ENDP
;
;*******************************************************************
;       implementation of x**y by exponent factoring algorithm 
;*******************************************************************
;***************************|
;     code segment data     |
;***************************|
X_TO_1000       DT      0
X_TO_100        DT      0
X_TO_10         DT      0
;
; Storage for entry values
X_VALUE         DT      0
;
X_TO_Y_BYFAC    PROC    FAR
; Exact calculation of x**y by exponent factoring according
; to the following factor list:
;
; F3 = 1000    I3 = INT(y/F3)            R3 = y MOD(I3*F3)
; F2 = 100     I2 = INT(R3/F2)           R2 = R3 MOD(I2*F2)
; F1 = 10      I1 = INT(R2/F1)           R1 = R2 MOD(I1*F1)
; F0 = 1       I0 = INT(R1/F0)
;
; Exponent factorization formula:
;
; 10**y = (10**F3)**I3 * (10**F2)**I2 * (10**F1)**I1 * (10**F0)**I0
;         ____________   ____________    ___________    ___________
;               |              |               |              |
;               |              |               |              |
;            STAGE 3        STAGE 2         STAGE 1       STAGE 0
; partial       |              |               |              |
; products     S3             S2              S1             S0 
;           ------------------------------------------------------
;                      e x e c u t i o n    s t a g e s
; On entry:
;         ST(0) = exponent
;         ST(1) = base
; On exit:
;         ST(0) = x**y
;
; During operations:
;         CS:X_VALUE holds base (extended precision real)
;         CS:Y_VALUE holds exponent (word integer)
;
;******************************|
;  move data to work variables |
;******************************|
;                               |    y     |     x    | -------- |
        FISTP   CS:Y_VALUE      ;    x     | -------- | -------- |
        FSTP    CS:X_VALUE      ; -------- |
;***************************|
;        special case       |
;       test for x**0       |
;***************************|
; By the rules of exponents any base to the 0 power = 1
        MOV     CX,CS:Y_VALUE   ; Exponent to loop register
        CMP     CX,0            ; Test for 0 exponent
        JNE     NOT_ZERO_EXP    ; Go if not zero
; At this point we have detected x**0
        FLD1                    ; Load 1
        RET
;***************************|
;        special case       |
;       test for x**0       |
;***************************|
; By the rules of exponents any base to the 0 power = 1
NOT_ZERO_EXP:
        MOV     CX,CS:Y_VALUE   ; Exponent to loop register
        CMP     CX,0            ; Test for 0 exponent
        JNE     NOT_EXP_0       ; Go if not zero
; At this point we have detected x**0
        FLD1                    ; Load 1
        RET
;****************************|
; calculate factors:         |
;             y**1000,       |
;             y**100,        |
;         and y**10          |
;****************************|
; These factors are needed in stages 3, 2, and 1 of the calculations
; Only the factors actually required are obtained
; Test for EXP > 1000
; First load 1 for the case that exp < 1000
NOT_EXP_0:
        FLD1                    ;    1    | ------- | ------- |
        CMP     CX,10           ; Is exp > 10
        JA      GET_FACTORS     ; Go if exponent > 10
; Load two more 1s for final product calculation
        FLD1
        FLD1                    ;    1    |    1    |    1    |
        JMP     STAGE_0         ; Go directly to last stage
;***************************|
;      calculate factors    |
;***************************|
; If the power is > 10 then all three factors are calculated and
; stored
GET_FACTORS:
        FSTP    ST(0)           ; ------- | ------- | ------- |
; Calculate y**1000. At this point CX is > 1000
        MOV     DX,CX           ; Save counter in DX
;****************************|
;      calculate x**1000     |
;        in 3 stages         |
;****************************|
; First stage: x**10
;                               |  ST(0)  |  ST(1)  |  ST(2)  |
        FLD     CS:X_VALUE      ;    x    | ------- | ------- |
        FLD1                    ;    1    |    x    | ------- |
        MOV     CX,10           ; Multiplier
GET_10X:
        FMUL    ST,ST(1)        ;  prod   |    x    | ------- |
        LOOP    GET_10X         ;  x**10  |    x    | ------- |
        FSTP    ST(1)           ;  x**10  | ------- | ------- |
        FLD     ST(0)           ;  x**10  |  x**10  | ------- |
; Store factor x**10
        FSTP    CS:X_TO_10
        FLD1                    ;   1     |  x**10  | ------- |
        MOV     CX,10           ; Multiplier for second loop
GET_100X:
        FMUL    ST,ST(1)        ;  prod   |  x**10  | ------- |
        LOOP    GET_100X        ;  x**100 |  x**10  | ------- |
        FSTP    ST(1)           ;  x**100 | ------- | ------- |
        FLD     ST(0)           ;  x**100 | x**100  |  x**10  |
; Store factor x**100
        FSTP    CS:X_TO_100
; At this point ST(0) = x**100
        FLD1                    ;   1     |  x**100 | ------- |
        MOV     CX,10           ; Multiplier for third loop
GET_1000X:
        FMUL    ST,ST(1)        ;  prod   |  x**100 | ------- |
        LOOP    GET_1000X       ; x**1000 |  x**100 | ------- |
        FSTP    ST(1)           ;  x**100 | ------- | ------- |
; Store factor x**1000
        FSTP    CS:X_TO_1000
;                               | ------- | ------- | ------- |
; All three factors are stored. Stack is empty
        MOV     CX,DX           ; Exponent to CX
;
;*****************|
;    STAGE 3      |
;********************************|
;    calculate (X**F3)**I3       |
;********************************|
; Test for EXP > 1000
; First load 1 for the case that exp < 1000
        FLD1                    ;    1    | ------- | ------- |
        CMP     CX,1000         ; Is exp > 1000
        JNA     STAGE_2         ; Go if not above than 1000
        FSTP    ST(0)           ; ------- | ------- | ------- |
; At this point CX is > 1000
; Load 1.0 E1000 in ST(0)
        FLD     CS:X_TO_1000    ;   S3    | ------- | ------- |
        FLD1                    ;   1     |    S3   | ------- |
;**************************|
; calculate INT(y/1000)    |
;**************************|
; The division by 1000 is performed in 2 steps in order to avoid
; a divide overflow
        MOV     DX,CX           ; Save original exponent in DX
        MOV     AX,CX           ; Exponent to AX
        MOV     BX,100          ; First divisor
        DIV     BL              ; Exponent / 100
        MOV     AH,0            ; Clear remainder of byte division
        MOV     BX,10           ; Second divisor
        DIV     BL              ; First dividend / 10
        MOV     AH,0            ; Clear remainder of byte division
        MOV     CX,AX           ; Copy divisor to CX
        MOV     BX,1000         ; Base for this stage
;*****************************|
; calculate  (x**F3)**I3      |
;*****************************|
; At this point CX = y MOD 100
MUL_STAGE_3:
        FMUL    ST,ST(1)        ;  prod   |    S3   | ------- |
        SUB     DX,BX           ; Exponent - 1000 to calculate
                                ; remainder
        LOOP    MUL_STAGE_3
        FSTP    ST(1)           ;   S3    | ------- | ------- |
; DX holds REM (y MOD 1000)
        MOV     CX,DX           ; Copy remainder to CX
; Test for exponent exact multiple of 1000 (no remainder)
        CMP     CX,0            ; In this case CX = 0
        JNE     STAGE_2         ; Go if not
;*************************|
;  quick exit if REM = 0  |
;***************************************************|
;   multiply exponent factors:                      |
;       (y**F3)**I3 -------> S3 --------->   ST(3)  |
;***************************************************|
;                             |    S3     | -------- | -------- |
        RET                   ; Return if exponent is exact
                              ; multiple of 1000
;
;*****************|
;    STAGE 2      |
;********************************|
;    calculate (x**F2)**I2       |
;********************************|
; Test for EXP > 100
; First load 1 for the case that exp < 100
; At this point:                |  ST(0)  |  ST(1)  | ST(2)  |
;                               |   S3    | ------- |
;                               |  or 1   |
STAGE_2:
; Test for EXP > 100
; First load 1 for the case that exp < 100
        FLD1                    ;    1    |   S3    | ------- |
        CMP     CX,100          ; Is exp > 100
        JNA     STAGE_1         ; Go if not above than 100
        FSTP    ST(0)           ; ------- | ------- | ------- |
; At this point CX is > 100
; Load x**100 factor
        FLD     CS:X_TO_100     ; x**100  |   S3    | ------- |
        FLD1                    ;   1     | x**100  |    S3   |
;*************************|
; calculate INT(y/100)    |
;*************************|
        MOV     DX,CX           ; Save original exponent in DX
        MOV     AX,CX           ; Exponent to AX
        MOV     BX,100          ; Divisor
        DIV     BL              ; Exponent / 100
        MOV     AH,0            ; Clear reminder
        MOV     CX,AX           ; Copy divisor to CX
;***************************|
; calculate (y**F2)**I2     |
;***************************|
; At this point CX = y MOD 100
MUL_STAGE_2:
        FMUL    ST,ST(1)        ;  prod   | x**100  |   S3    |
        SUB     DX,BX           ; Exponent - 100 to calculate
                                ; remainder
        LOOP    MUL_STAGE_2
        FSTP    ST(1)           ;    S2   |    S3   | ------- |
; DX holds REM (y MOD 100)
        MOV     CX,DX           ; Copy remainder to CX
; Test for exponent exact multiple of 100 (no remainder)
        CMP     CX,0            ; In this case CX = 0
        JNE     STAGE_1         ; Go if not
;*************************|
;  quick exit if REM = 0  |
;***************************************************|
;   multiply exponent factors:                      |
;       (y**F3)**I3 -------> S3 --------->   ST(3)  |
;       (y**F2)**I2 -------> S2 --------->   ST(2)  |
;***************************************************|
;                             |    S2     |    S3    | -------- |
        FMUL    ST,ST(1)
        FSTP    ST(1)         ;   x**y    | ------- | -------- |
        RET                   ; Exponent is exact multiple of 100
;
;*****************|
;    STAGE 1      |
;********************************|
;    calculate (x**F1)**I1       |
;********************************|
; Test for EXP > 10
; First load 1 for the case that exp < 10
; At this point:                |  ST(0)  |  ST(1)  |  ST(2)  |
;                               |   S2    |   S3    | ------- |
;                               |  or 1   |   or 1
; CX = REM (y MOD 100)
; Test for EXP > 10
; First load 1 for the case that exp < 10 
STAGE_1:
        FLD1                    ;    1    |   S2    |    S3   |
        CMP     CX,10           ; Is exp > 10 
        JNA     STAGE_0         ; Go if not above than 10
        FSTP    ST(0)           ;    S2   |   S3    | ------- |
; At this point CX is > 10
        FLD     CS:X_TO_10      ; x**10   |   S2    |    S3   |
        FLD1                    ;   1     |  x**10  |    S2   |..
;*******************************|
;  calculate (x**F1)**I1        |
;*******************************|
; CX = REM (y MOD 100)
        MOV     DX,CX           ; Save remainder in DX
        MOV     AX,CX           ; New exponent to AX
        MOV     BX,10           ; Divisor
        DIV     BL              ; Exponent / 25 
        MOV     AH,0            ; Clear reminder
        MOV     CX,AX           ; Copy divisor to CX
; CX = INT(y/100)
MUL_STAGE_1:
        FMUL    ST,ST(1)        ;  prod   | x**10   |   S2    |..
        SUB     DX,BX           ; Exponent - 10 to calculate
                                ; remainder
        LOOP    MUL_STAGE_1
        FSTP    ST(1)           ;   S1    |   S2    |   S3    |
; DX holds remainder
        MOV     CX,DX           ; CX is new iteration counter
; Test for exponent exact multiple of 10
        CMP     CX,0            ; In this case CX = 0
        JNE     STAGE_0         ; Go if not
;*************************|
;  quick exit if REM = 0  |
;***************************************************|
;   multiply exponent factors:                      |
;       (y**F3)**I3 -------> S3 --------->   ST(3)  |
;       (y**F2)**I2 -------> S2 --------->   ST(2)  |
;       (y**F1)**I1 -------> S1 --------->   ST(1)  |
;***************************************************|
;                             |    S1     |    S2    |    S3    |..
        FMUL    ST,ST(1)
        FSTP    ST(1)         ;    PP     |    S3    | -------- |
        FMUL    ST,ST(1)
        FSTP    ST(1)         ;   x**y    | -------- | -------- |
        RET
;
;*****************|
;     STAGE 0     |
;********************************|
;    calculate (y**F0)**I0       |
;********************************|
; At this point:                |  ST(0)  |  ST(1)  |  ST(2)  |
;                               |   S1    |   S2    |   S3    |
;                               |  or 1   |   or 1  |   or 1  |
; CX = INT(y/10)
STAGE_0:
        FLD     CS:X_VALUE      ;    x    |    S1   |   S2    |..
        FLD1                    ;    1    |    x    |   S2    |..
MUL_STAGE_0:
        FMUL    ST,ST(1)        ;   S0    |    x    |   S1    |..
        LOOP    MUL_STAGE_0
        FSTP    ST(1)           ;   S0    |   S1    |   S2    |..
;***************************************************|
;   multiply exponent factors:                      |
;       (y**F3)**I3 -------> S3 --------->   ST(3)  |
;       (y**F2)**I2 -------> S2 --------->   ST(2)  |
;       (y**F1)**I1 -------> S1 --------->   ST(1)  |
;       (y**F0)**I0 -------> S0 --------->   ST(0)  |
;***************************************************|
;                            |     S0     |    S1    |    S2    |..
        FMUL    ST,ST(1)
        FSTP    ST(1)        ;     PP     |    S2    |    S3    |
        FMUL    ST,ST(1)
        FSTP    ST(1)        ;     PP     |    S3    | -------- |
        FMUL    ST,ST(1)
        FSTP    ST(1)        ;   x**y     | ------- | -------- |
        RET
;
X_TO_Y_BYFAC    ENDP

;********************************************************************
;         implementation of x**y by binary powering algorithm 
;********************************************************************
;***************************|
;     code segment data     |
;***************************|
OP_COUNT        DB      0       ; Iterations counter
;
X_TO_Y_BYBP     PROC    FAR
; Calculation of x**y by binary powering
; On entry:
;         ST(0) = exponent
;         ST(1) = base
; On exit:
;         ST(0) = x**y
;
; Algorithm (based on D. Knuth)
;      1. Determine maximum number of binary digits in
;         exponent. This is intial value of operations counter
;      2. Skip leading zeros in exponent decrementing operations
;         counter
;      3. Skip first 1-digit decrementing operations counter
;      4. Test leftmost binary digit of exponent
;         If 1
;             Square previous product and multiply by x
;             Goto step 5
;         If 0
;             Square previous product
;      5. Shift left exponent bits
;         Decrement operations counter
;             If not zero then step 4
;      6. End
;
;*****************************|
; move data to work variable  |
;   init operations counter   |
;*****************************|
;                               |    y     |     x    | -------- |
        FISTP   CS:Y_VALUE      ;    x     | -------- | -------- |
        MOV     DX,CS:Y_VALUE       ; Exponent to loop register
        MOV     CS:OP_COUNT,16  ; Initialize operations counter
; DX hold exponent
;***************************|
;        special case       |
;       test for x**0       |
;***************************|
; By the rules of exponents any base to the 0 power = 1
        CMP     DX,0            ; Test for 0 exponent
        JNE     TEST_CASE_E1    ; Go if not zero
; At this point we have detected x**0
        FSTP    ST(0)           ; Clear stack
        FLD1                    ; Load 1
        RET
;***************************|
;        special case       |
;       test for x**1       |
;***************************|
; By the rules of exponents any base to the 1 power = base
TEST_CASE_E1:
        CMP     DX,1            ; Test for 1 exponent
        JNE     SUPRESS_0S      ; Go if not zero
; At this point we have detected x**1
; Base is already in ST(0)
;                               |    x     | -------- | -------- |
        RET
;****************************|
;     suppress leading 0s    |
;   decrementing ops counter |
;****************************|
SUPRESS_0S:
        TEST    DX,8000H        ; Is bit 15 set?
        JNZ     SUPRESS_LAST    ; Go if set
; Bit 15 is not set
        DEC     CS:OP_COUNT     ; Adjust shift counter
        SAL     DX,1            ; Shift bits left
        JMP     SUPRESS_0S
; At this point all leading 0 bits have been eliminated
; OP_COUNT has been decremented accordingly
;***************************|
;    skip first 1-bit       |
;   decrement ops counter   |
;***************************|
SUPRESS_LAST:
        SAL     DX,1            ; Shift out this 1-bit
        DEC     CS:OP_COUNT     ; Adjust shift counter
        FLD     ST(0)           ;    x     |     x    | -------- |
;***************************|
;     test leftmost bit     |
; if 1, square and mult.    |
; if 0, square only         |
;***************************|
TEST_BIT:
        TEST    DX,8000H        ; Is bit 15 set?
        JNZ     SQUARE_AND_MUL10        ; Go if set
; Square only
        FMUL    ST,ST(0)        ; ST**2    |    10    | -------- |
        JMP     NEXT_OP
SQUARE_AND_MUL10:
        FMUL    ST,ST(0)        ; ST**2    |    10    | -------- |
        FMUL    ST,ST(1)        ; ST*10    |    10    | -------- |
;***************************|
;  shift exponent bits and  |
;  test for end of counter  |
;***************************|
NEXT_OP:
        SAL     DX,1            ; Shift all bits left
        DEC     CS:OP_COUNT
; Test for end of processing
        JNZ     TEST_BIT        ; Continue if not zero
;***************************|
;    end of routine         |
;***************************|
        FSTP    ST(1)           ;  x**y    | -------- |
        RET
X_TO_Y_BYBP     ENDP
;
;********************************************************************
;    implementation of x**y by logarithmic approximation algorithm 
;********************************************************************
; Constant defined in single precision format
ONE_HALF        DD      0.5     ; 1/2 in single precision
;***************************|
;     code segment data     |
;***************************|
STATUS_WW       DW      0       ; Machine status word
CONTROL_WW      DW      0       ; Control word
ROUND_DOWN      DW      177FH   ; Control word for round down
                                ; operations
;
X_TO_Y_BYLOG    PROC    FAR
; Raise a nymber (y) to an arbitrary power (x)
; On entry:
;         ST(0) = exponent
;         ST(1) = base
; On exit:
;         ST(0) = x**y
; 
; The FPU exponential function x**y is implemented by the following
; identity:
;            x**y = 2**(y*Log2(x))

; Algorithm to raise 2 to an arbitrary power::
;            2**x = (2**i)*(2**f)
;            where:
;                  i = the nearest integer smaller than x
;                  f = the fraction x - i
; In calculations:
;            if f > 1/2 then
;               2**f = 2**((f-1/2) + 1/2)
;                    = 2**(f-1/2) * 2**(1/2)
;                    = 2**(f-1/2) * (square root of 2)
; This manipulation allows reducing f to the range of the instruction
; F2XM1. 2**i is calculated using FSCALE
;
        FXCH                    ; Invert x and y
;                       |    ST(0)    |    ST(1)    |   ST(2)
                        ;      y      |      x      | --------- |
        FYL2X           ;   y*LOG2(x) | ----------- | --------- |
;**********************|
;   set to round down  |
;**********************|
        FSTCW   CS:CONTROL_WW   ; Store control word in memory
        FLDCW   CS:ROUND_DOWN   ; Install new control word
; At this point the 80x87 is set to round down
; This ensures that i is smaller than x
;                       |    ST(0)    |    ST(1)   |    ST(2)
                        ;      x      | ---------- | --------- |
        FLD     ST(0)   ;      x      |     x      | --------- |
        FRNDINT         ;      i      |     x      | --------- |
        FLDCW   CS:CONTROL_WW   ; Restore original control word
        FSUB    ST(1),ST;      i      |     f      | --------- |
; f = x - i, that is, the fractional part of x
        FXCH            ;      f      |     i      | --------- |
        FLD     CS:ONE_HALF
                        ;     1/2     |     f      |     i     |
        FXCH            ;      f      |    1/2     |     i     |
; Scale f to the range of F2XM1, which is 0 <= x => 1/2
        FPREM           ; REM(f/(1/2) |    1/2     |     i
; If the value of f => 0.5 then the quotient of FPREM is 1
; and bit C1 is set. Otherwise x is < 1/2
        FSTSW   CS:STATUS_WW    ; Store status word
        FWAIT
;                       |      f      |     1/2    |     i    |
        FSTP    ST(1)   ;      f      |     i      | -------- |
        F2XM1           ;   2**(f)-1  |     i      | -------- |
; Since F2XM1 calculates 2^(f)-1, 1 must be added to the
; result
        FLD1            ;      1      | 2**(f)-1   |     i   |
        FADDP   ST(1),ST;    2**f     |     i      | ------- |
        MOV     AH,BYTE PTR CS:STATUS_WW+1
                        ; Move status bits into AH
        TEST    AH,00000010B    ; Test bit C1
        JZ      CASE_2
; If f was => 1/2 then multiply by square root of 2
        FLD1            ;     1       |   2**f     |     i   |
        FADD    ST,ST(0);     2       |   2**f     |     i   |
        FSQRT           ;  2**(1/2)   |   2**f     |     i   |
        FMULP   ST(1),ST;2**f*2**(1/2)|     i      | ------- |
CASE_2:
;                       |    2**f     |      i     | ------- |
        FSCALE          ;  2(f)*2(i)  |      i     | ------- |
        FSTP    ST(1)   ;    2**x     | ---------- |
        RET
X_TO_Y_BYLOG    ENDP
;
;****************************************************************
;        mixed x**y (uses both logs and binary powering)
;****************************************************************
;***************************|
;    code segment storage   |
;***************************|
; Temporary storage for exponent and base
EXP_ELEMENT    DT      0       ; Storage for exponent
BASE_ELEMENT   DT      0       ; Storage for base
;
X_TO_Y_MIXED    PROC    FAR
; Mixed method of obtaining x**y
;    Integer part is calculated by binary powering
;    Fractional part is calculated by log approximation
; On entry:
;         ST(0) = exponent
;         ST(1) = base (must be integer)
; On exit:
;         ST(0) = x**y
;
; This routine calls the procedures named X_TO_Y_BYBP and
; X_TO_Y_BYLOG
;
; First calculate integer part using binary powering
                                ;   exp    |   base   |
        FSTP    CS:EXP_ELEMENT  ;   base   | -------- |
        FSTP    CS:BASE_ELEMENT ; -------- |
        FLD     CS:EXP_ELEMENT  ;   exp    |   base   |
        FRNDINT                 ; int(exp) |   base   |
        FLD     CS:BASE_ELEMENT ;  base    | int(exp) |
        FXCH                    ; int(exp) |   base   |
        CALL    FAR PTR X_TO_Y_BYBP
                                ; x** int y| -------- |
; Now calculate fractional part using log approximation
        FLD     CS:EXP_ELEMENT  ;   exp    | x**int y|
        FLD     ST(0)           ;   exp    |   exp   | x**int y |
        FRNDINT                 ; int(exp) |   exp   | x**int y |
        FSUBP   ST(1),ST        ; frc(exp) | x**int y| -------- |
        FLD     CS:BASE_ELEMENT ;  base    | frc(exp)| x**int y |
        FXCH                    ; frc(exp) |   base  | x**int y |
        CALL    FAR PTR X_TO_Y_BYLOG
                                ;x**(frc y)| x**int y| -------- |
        FMUL    ST,ST(1)        ;  x**y    | x**int y| -------- |
        FSTP    ST(1)           ;  x**y    | --------|
; ST(0) holds x**y
        RET
X_TO_Y_MIXED    ENDP
;*****************************************************************
;                            logarithms
;*****************************************************************
LOG_E           PROC    FAR
; Calculate the natural log (base e)
; On entry:
;         ST(0) = positive number
; On exit:
;         ST(0) = ln ST(0) (Natural log)
;
;                       |    ST(0)    |    ST(1)   |    ST(2)
                        ;      x      | ---------- | ---------- |
        FLDLN2          ;   LOGe(2)   |     x      | ---------- |
        FXCH            ;      x      |   LOGe(2)  | ---------- |
        FYL2X           ;  LOGe(2)*   | 
                        ;   LOG2(x)   | ---------- |
        RET
LOG_E           ENDP
;*****************************************************************
;
LOG_10          PROC    FAR
; Calculate the common log (Base 10)
; On entry:
;         ST(0) = positive number
; On exit:
;         ST(0) = LOG ST(0) (Common log)
;
;                       |    ST(0)    |    ST(1)   |    ST(2)
                        ;      x      | ---------- | ---------- |
        FLDLG2          ;   LOG10(2)  |     x      | ---------- |
        FXCH            ;      x      |   LOG10(2) | ---------- |
        FYL2X           ;  LOG10(2)*  |
                        ;   LOG2(x)   | ---------- |
        RET
;
LOG_10          ENDP
;*****************************************************************
;
ALOG_10         PROC    FAR
; Calculate the common antilogarithm of the number in ST(0)
; On entry:
;         ST(0) holds a positive logarithm base 10
; On exit:
;         ST(0) holds the antilogarithm of the number
;
;                       |    ST(0)    |    ST(1)   |    ST(2)
        FLDL2T          ;   log2(10)  |      x     | ---------- |
        CALL    FAR PTR ALOG_2
                        ;  alog10(x)  | ---------- |
        RET
ALOG_10         ENDP
;*****************************************************************
;
ALOG_E          PROC    FAR
; Calculate the common antilogarithm of the number in ST(0)
; On entry:
;         ST(0) holds a positive logarithm base 10
; On exit:
;         ST(0) holds the antilogarithm of the number
;
;                       |    ST(0)    |    ST(1)   |    ST(2)
        FLDL2E          ;   log2(10)  |      x     | ---------- |
        CALL    FAR PTR ALOG_2
                        ; alog2(x)    | ---------- |
        RET
ALOG_E          ENDP
;*****************************************************************
;
ALOG_2          PROC    FAR
; Core routine to calculate antilog base e and base 10
; On entry:
;         ST(0) = logarithm base 2 of a number (N)
;         ST(1) = positive power to which N must be raised (x)
; On exit:
;         ST(0) = N**x

;
;                       |    ST(0)    |    ST(1)   |    ST(2)
;                       |  log2(N)    |      x     | --------- |
        FMULP   ST(1),ST; LOG2(N)*x=z |    empty
; Set control word to round down
        FSTCW   CS:CONTROL_WW   ; Store caller's control word
        FLDCW   CS:ROUND_DOWN   ; Temporary control word for
                                ; rounding down
        FLD1            ;     1       |      z     | --------- |
        FCHS            ;    -1       |      z     | --------- |
        FLD     ST(1)   ;     z       |     -1     |     z     |
        FRNDINT         ; i = int(z)  |     -1     |     z     |
; The value i is the closest integer smaller than z
        FLDCW   CS:CONTROL_WW   ; Restore caller's control word
        FXCH    ST(2)   ;     z       |     -1     |     i     |
        FSUB    ST,ST(2);     f       |     -1     |     i     |
        FSCALE          ;    f/2      |     -1     |     i     |
; FSCALE using the value -1 decrements the binary exponent of
; ST(0), which is the same as dividing by 2
        F2XM1           ;2**(f/2) -1  |     -1     |     i     |
        FSTP    ST(1)   ;2**(f/2)- 1  |      i     | --------- |
        FLD1            ;      1      |2**(f/2) - 1|     i     |
        FADDP   ST(1),ST;  2**(f/2)   |      i     | --------- |
        FMUL    ST,ST(0);    2**f     |      i     | --------- |
        FSCALE          ; 2**f * 2**i |      i     | --------- |
        FSTP    ST(1)   ;     N**x    | ---------- |
        RET
ALOG_2          ENDP
;
;*****************************************************************
;                  input and output super procedures
;*****************************************************************
;***************************|
;   code segment data       |
;***************************|
DOS_BUF         DB      28      ; Maximum input
                DB      0       ; Actual input
DOS_STRING      DB      30 DUP (0H)     ; Buffer space
EXP_STRING      DB      28 DUP (0H)
                DB      '$'
;
INPUT_NUMBER    PROC    FAR
; Procedure to enter a number into the FPU stack top register
        PUSH    DS              ; Save caller's
        PUSH    CS              ; CS to stack
        POP     DS              ; DS = CS
        LEA     DX,CS:DOS_BUF   ; Local input buffer for 28
                                ; characters
        MOV     AH,10           ; DOS service request number
        INT     21H             ; Transfer to DOS
; At this point input string is in DOS_STRING buffer
; Convert string to exponential form
        LEA     SI,CS:DOS_STRING   ; Pointer to source string
        LEA     DI,CS:EXP_STRING   ; Pointer to destination string
        CALL    ASCII_TO_EXP
; Input string is now in exponential form
        LEA     DX,CS:EXP_STRING   ; Pointer to ASCII string
        CALL    FPU_INPUT
; At this point input is in ST(0)
; New line
        LEA     DX,NEW_LINE
        MOV     AH,9
        INT     21H
; Restore caller's DS
        POP     DS
        RET
INPUT_NUMBER    ENDP
;*****************************************************************
;***************************|
;   code segment data       |
;***************************|
ASCII_NUM       DB      27 DUP (0H)
                DB      '$'
NEW_LINE        DB      0AH,0DH,'$'
;
OUTPUT_NUMBER   PROC    FAR
; Procedure to display a number in FPU stack top
        PUSH    DS              ; Save caller's
        PUSH    CS              ; CS to stack
        POP     DS              ; DS = CS
; Convert ST(0) to ASCII exponential form
        LEA     DX,CS:ASCII_NUM    ; Buffer for storage
        CALL    FPU_OUTPUT
; Use MS DOS service number 9 to display converted number
        LEA     DX,CS:ASCII_NUM ; Pointer to number string
        MOV     AH,9            ; DOS service request
        INT     21H
; New line
        LEA     DX,NEW_LINE
        MOV     AH,9
        INT     21H
        POP     DS              ; Restore caller's DS
        RET
OUTPUT_NUMBER   ENDP

CODE    ENDS
        END

 21H
        POP    